import random
import time

from django.core.management.base import BaseCommand
from faker import Factory

from ....acl.cache import clear_acl_cache
from ....categories.models import Category
from ....core.management.progressbar import show_progress
from ...categories import fake_category, fake_closed_category


class Command(BaseCommand):
    help = "Creates fake categories for dev and testing purposes."

    def add_arguments(self, parser):
        parser.add_argument(
            "categories",
            help="number of categories to create",
            nargs="?",
            type=int,
            default=5,
        )

        parser.add_argument(
            "level",
            help="level of created categories",
            nargs="?",
            type=int,
            default=1,
        )

    def handle(self, *args, **options):  # pylint: disable=too-many-locals
        items_to_create = options["categories"]
        level = options["level"]

        fake = Factory.create()

        categories = Category.objects.all_categories(include_root=True).filter(
            level=level - 1
        )
        acl_source = list(Category.objects.all_categories())[0]

        if not categories.exists():
            self.stdout.write("No valid parent categories exist.\n")
            return

        message = "Creating %s fake categories...\n"
        self.stdout.write(message % items_to_create)

        created_count = 0
        start_time = time.time()
        show_progress(self, created_count, items_to_create)

        while created_count < items_to_create:
            parent = random.choice(categories)

            if random.randint(0, 100) > 90:
                fake_closed_category(fake, parent, copy_acl_from=acl_source)
            else:
                fake_category(fake, parent, copy_acl_from=acl_source)

            created_count += 1
            show_progress(self, created_count, items_to_create, start_time)

        clear_acl_cache()

        total_time = time.time() - start_time
        total_humanized = time.strftime("%H:%M:%S", time.gmtime(total_time))
        message = "\n\nSuccessfully created %s fake categories in %s"
        self.stdout.write(message % (created_count, total_humanized))



from django.test import TestCase
from django.urls import reverse

from ..models import Agreement


class PrivacyPolicyTests(TestCase):
    def setUp(self):
        Agreement.objects.invalidate_cache()

    def tearDown(self):
        Agreement.objects.invalidate_cache()

    def test_404_on_no_policy(self):
        """policy view returns 404 when no policy is set"""
        response = self.client.get(reverse("misago:privacy-policy"))
        self.assertEqual(response.status_code, 404)

    def test_301_on_link_policy(self):
        """policy view returns 302 redirect when link is set"""
        Agreement.objects.create(
            type=Agreement.TYPE_PRIVACY,
            link="http://test.com",
            text="Lorem ipsum",
            is_active=True,
        )

        response = self.client.get(reverse("misago:privacy-policy"))
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response["location"], "http://test.com")

    def test_200_on_link_policy(self):
        """policy view returns 200 when custom tos content is set"""
        Agreement.objects.create(
            type=Agreement.TYPE_PRIVACY,
            title="Test Policy",
            text="Lorem ipsum dolor",
            is_active=True,
        )

        response = self.client.get(reverse("misago:privacy-policy"))
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, "Test Policy")
        self.assertContains(response, "Lorem ipsum dolor")


class TermsOfServiceTests(TestCase):
    def setUp(self):
        Agreement.objects.invalidate_cache()

    def tearDown(self):
        Agreement.objects.invalidate_cache()

    def test_404_on_no_tos(self):
        """TOS view returns 404 when no TOS is set"""
        response = self.client.get(reverse("misago:terms-of-service"))
        self.assertEqual(response.status_code, 404)

    def test_301_on_link_tos(self):
        """TOS view returns 302 redirect when link is set"""
        Agreement.objects.create(
            type=Agreement.TYPE_TOS,
            link="http://test.com",
            text="Lorem ipsum",
            is_active=True,
        )

        response = self.client.get(reverse("misago:terms-of-service"))
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response["location"], "http://test.com")

    def test_200_on_link_tos(self):
        """TOS view returns 200 when custom tos content is set"""
        Agreement.objects.create(
            type=Agreement.TYPE_TOS,
            title="Test ToS",
            text="Lorem ipsum dolor",
            is_active=True,
        )

        response = self.client.get(reverse("misago:terms-of-service"))
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, "Test ToS")
        self.assertContains(response, "Lorem ipsum dolor")



from django.db import transaction
from django.utils import timezone

from ...notifications.tasks import delete_duplicate_watched_threads
from ..events import record_event

__all__ = [
    "change_thread_title",
    "pin_thread_globally",
    "pin_thread_locally",
    "unpin_thread",
    "move_thread",
    "merge_thread",
    "approve_thread",
    "open_thread",
    "close_thread",
    "unhide_thread",
    "hide_thread",
    "delete_thread",
]


@transaction.atomic
def change_thread_title(request, thread, new_title):
    if thread.title == new_title:
        return False

    old_title = thread.title
    thread.set_title(new_title)
    thread.save(update_fields=["title", "slug"])

    thread.first_post.set_search_document(thread.title)
    thread.first_post.save(update_fields=["search_document"])

    thread.first_post.update_search_vector()
    thread.first_post.save(update_fields=["search_vector"])

    record_event(request, thread, "changed_title", {"old_title": old_title})
    return True


@transaction.atomic
def pin_thread_globally(request, thread):
    if thread.weight == 2:
        return False

    thread.weight = 2
    record_event(request, thread, "pinned_globally")
    return True


@transaction.atomic
def pin_thread_locally(request, thread):
    if thread.weight == 1:
        return False

    thread.weight = 1
    record_event(request, thread, "pinned_locally")
    return True


@transaction.atomic
def unpin_thread(request, thread):
    if thread.weight == 0:
        return False

    thread.weight = 0
    record_event(request, thread, "unpinned")
    return True


@transaction.atomic
def move_thread(request, thread, new_category):
    if thread.category_id == new_category.pk:
        return False

    from_category = thread.category
    thread.move(new_category)

    record_event(
        request,
        thread,
        "moved",
        {
            "from_category": {
                "name": from_category.name,
                "url": from_category.get_absolute_url(),
            }
        },
    )
    return True


@transaction.atomic
def merge_thread(request, thread, other_thread):
    thread.merge(other_thread)
    other_thread.delete()

    delete_duplicate_watched_threads.delay(thread.id)

    record_event(request, thread, "merged", {"merged_thread": other_thread.title})
    return True


@transaction.atomic
def approve_thread(request, thread):
    if not thread.is_unapproved:
        return False

    thread.first_post.is_unapproved = False
    thread.first_post.save(update_fields=["is_unapproved"])

    thread.is_unapproved = False

    unapproved_post_qs = thread.post_set.filter(is_unapproved=True)
    thread.has_unapproved_posts = unapproved_post_qs.exists()

    record_event(request, thread, "approved")
    return True


@transaction.atomic
def open_thread(request, thread):
    if not thread.is_closed:
        return False

    thread.is_closed = False
    record_event(request, thread, "opened")
    return True


@transaction.atomic
def close_thread(request, thread):
    if thread.is_closed:
        return False

    thread.is_closed = True
    record_event(request, thread, "closed")
    return True


@transaction.atomic
def unhide_thread(request, thread):
    if not thread.is_hidden:
        return False

    thread.first_post.is_hidden = False
    thread.first_post.save(update_fields=["is_hidden"])
    thread.is_hidden = False

    record_event(request, thread, "unhid")

    if thread.pk == thread.category.last_thread_id:
        thread.category.synchronize()
        thread.category.save()

    return True


@transaction.atomic
def hide_thread(request, thread):
    if thread.is_hidden:
        return False

    thread.first_post.is_hidden = True
    thread.first_post.hidden_by = request.user
    thread.first_post.hidden_by_name = request.user.username
    thread.first_post.hidden_by_slug = request.user.slug
    thread.first_post.hidden_on = timezone.now()
    thread.first_post.save(
        update_fields=[
            "is_hidden",
            "hidden_by",
            "hidden_by_name",
            "hidden_by_slug",
            "hidden_on",
        ]
    )
    thread.is_hidden = True

    record_event(request, thread, "hid")

    if thread.pk == thread.category.last_thread_id:
        thread.category.synchronize()
        thread.category.save()

    return True


@transaction.atomic
def delete_thread(request, thread):
    thread.delete()

    thread.category.synchronize()
    thread.category.save()

    return True



import re
from dataclasses import dataclass
from importlib import import_module
from pathlib import Path
from types import ModuleType
from typing import Dict, List, Optional
from urllib.parse import urlparse

from django.conf import settings
from django.utils.translation import pgettext_lazy

from .manifest import MisagoPlugin


@dataclass(frozen=True)
class PluginMetadataUrl:
    url: str
    netloc: str


@dataclass(frozen=True)
class PluginMetadata:
    package: str
    dirname: str
    has_manifest: bool
    manifest_error: Optional[str]

    name: Optional[str]
    description: Optional[str]
    license: Optional[str]
    icon: Optional[str]
    color: Optional[str]
    version: Optional[str]
    author: Optional[str]
    homepage: Optional[PluginMetadataUrl]
    sponsor: Optional[PluginMetadataUrl]
    help: Optional[PluginMetadataUrl]
    bugs: Optional[PluginMetadataUrl]
    repo: Optional[PluginMetadataUrl]

    @property
    def has_urls(self) -> bool:
        return bool(
            self.homepage or self.sponsor or self.help or self.bugs or self.repo
        )


class PluginsMetadataLoader:
    initialized: bool
    plugins: List[str]
    metadata: Dict[str, PluginMetadata]

    def __init__(self, plugins: List[str]):
        self.initialized = False
        self.plugins = plugins
        self.metadata = {}

    def get_metadata(self) -> Dict[str, PluginMetadata]:
        if not self.initialized:
            self.metadata = self.build_metadata(self.plugins)
            self.initialized = True

        return self.metadata

    def build_metadata(self, plugins: List[str]) -> Dict[str, PluginMetadata]:
        metadata: Dict[str, PluginMetadata] = {}
        for plugin_package in plugins:
            metadata[plugin_package] = self.build_plugin_metadata(plugin_package)
        return metadata

    def build_plugin_metadata(self, plugin_package: str) -> PluginMetadata:
        package_obj = import_module(plugin_package)

        manifest: Optional[MisagoPlugin] = None
        manifest_error: Optional[str] = None

        try:
            manifest_module = import_module(f"{plugin_package}.misago_plugin")
            manifest_attr = getattr(manifest_module, "manifest", None)
            if manifest_attr is not None and not isinstance(
                manifest_attr, MisagoPlugin
            ):
                raise TypeError(
                    pgettext_lazy(
                        "plugins metadata",
                        "Plugin manifest is of an unexpected type '%(manifest_type)s' (expected MisagoPlugin instance)",
                    )
                    % {"manifest_type": type(manifest_attr).__name__}
                )
            manifest = manifest_attr
        except Exception as exc:
            manifest_error = f"{type(exc).__name__}: {exc}"

        return create_plugin_metadata(
            plugin_package,
            package_obj,
            manifest,
            manifest_error,
        )


def create_plugin_metadata(
    plugin_package: str,
    package_obj: ModuleType,
    manifest: Optional[MisagoPlugin],
    manifest_error: Optional[str],
) -> PluginMetadata:
    has_manifest = False

    name: Optional[str] = None
    description: Optional[str] = None
    license: Optional[str] = None
    icon: Optional[str] = None
    color: Optional[str] = None
    version: Optional[str] = None
    author: Optional[str] = None
    homepage: Optional[PluginMetadataUrl] = None
    sponsor: Optional[PluginMetadataUrl] = None
    help: Optional[PluginMetadataUrl] = None
    bugs: Optional[PluginMetadataUrl] = None
    repo: Optional[PluginMetadataUrl] = None

    if manifest:
        has_manifest = True
        name = clean_plugin_str(manifest.name, 100)
        description = clean_plugin_str(manifest.description, 250)
        license = clean_plugin_str(manifest.license, 50)
        icon = clean_plugin_icon(manifest.icon)
        color = clean_plugin_color(manifest.color)
        version = clean_plugin_str(manifest.version, 50)
        author = clean_plugin_str(manifest.author, 150)
        homepage = clean_plugin_url(manifest.homepage)
        sponsor = clean_plugin_url(manifest.sponsor)
        help = clean_plugin_url(manifest.help)
        bugs = clean_plugin_url(manifest.bugs)
        repo = clean_plugin_url(manifest.repo)

    return PluginMetadata(
        package=plugin_package,
        dirname=Path(package_obj.__file__).parent.parent.name,
        has_manifest=has_manifest,
        manifest_error=manifest_error,
        name=name,
        description=description,
        license=license,
        icon=icon,
        color=color,
        version=version,
        author=author,
        homepage=homepage,
        sponsor=sponsor,
        help=help,
        bugs=bugs,
        repo=repo,
    )


def clean_plugin_str(value: Optional[str], max_length: int) -> Optional[str]:
    if isinstance(value, str):
        return (value.strip()[:max_length].strip()) or None
    return None


URL_RE = re.compile("^https?://[A-Za-z0-9-_/?=&#.]+$")


def clean_plugin_url(value: Optional[str]) -> Optional[PluginMetadataUrl]:
    if isinstance(value, str) and URL_RE.match(value):
        try:
            url = urlparse(value)
            if not url:
                return None

            return PluginMetadataUrl(
                url=value,
                netloc=url.netloc,
            )
        except ValueError:
            return None
    return None


ICON_RE = re.compile("^fa(r|s)? fa-[a-z-]+$")


def clean_plugin_icon(value: Optional[str]) -> Optional[str]:
    if isinstance(value, str) and ICON_RE.match(value):
        return value
    return None


COLOR_RE = re.compile("^#[0-9a-f]{6}$", re.IGNORECASE)


def clean_plugin_color(value: Optional[str]) -> Optional[str]:
    if isinstance(value, str) and COLOR_RE.match(value):
        return value.upper()
    return None


plugins_metadata = PluginsMetadataLoader(settings.INSTALLED_PLUGINS)



# Generated by Django 2.2.1 on 2019-05-19 00:16

from django.conf import settings
from django.db import migrations

from ..hydrators import dehydrate_value

default_settings = [
    {"setting": "account_activation", "dry_value": "none", "is_public": True},
    {"setting": "allow_custom_avatars", "python_type": "bool", "wet_value": True},
    {
        "setting": "avatar_upload_limit",
        "python_type": "int",
        "wet_value": 1536,
        "is_public": True,
    },
    {"setting": "attachment_403_image", "python_type": "image"},
    {"setting": "attachment_404_image", "python_type": "image"},
    {"setting": "blank_avatar", "python_type": "image"},
    {"setting": "captcha_type", "dry_value": "no", "is_public": True},
    {"setting": "default_avatar", "dry_value": "gravatar"},
    {"setting": "default_gravatar_fallback", "dry_value": "dynamic"},
    {"setting": "unused_attachments_lifetime", "python_type": "int", "wet_value": 24},
    {"setting": "email_footer"},
    {
        "setting": "forum_address",
        "dry_value": getattr(settings, "MISAGO_ADDRESS", None),
        "is_public": True,
    },
    {"setting": "forum_footnote", "is_public": True},
    {"setting": "forum_name", "dry_value": "Misago", "is_public": True},
    {"setting": "google_tracking_id"},
    {"setting": "google_site_verification"},
    {"setting": "index_header", "is_public": True},
    {"setting": "index_meta_description", "is_public": True},
    {"setting": "index_title", "is_public": True},
    {"setting": "logo", "python_type": "image", "is_public": True},
    {"setting": "logo_small", "python_type": "image", "is_public": True},
    {"setting": "logo_text", "dry_value": "Misago", "is_public": True},
    {"setting": "daily_post_limit", "python_type": "int", "wet_value": 600},
    {"setting": "hourly_post_limit", "python_type": "int", "wet_value": 100},
    {"setting": "post_attachments_limit", "python_type": "int", "wet_value": 16},
    {
        "setting": "post_length_max",
        "python_type": "int",
        "wet_value": 60000,
        "is_public": True,
    },
    {
        "setting": "post_length_min",
        "python_type": "int",
        "wet_value": 5,
        "is_public": True,
    },
    {"setting": "readtracker_cutoff", "python_type": "int", "wet_value": 40},
    {"setting": "threads_per_page", "python_type": "int", "wet_value": 26},
    {"setting": "posts_per_page", "python_type": "int", "wet_value": 18},
    {"setting": "posts_per_page_orphans", "python_type": "int", "wet_value": 6},
    {"setting": "events_per_page", "python_type": "int", "wet_value": 20},
    {"setting": "og_image", "python_type": "image"},
    {
        "setting": "og_image_avatar_on_profile",
        "python_type": "bool",
        "wet_value": False,
    },
    {"setting": "og_image_avatar_on_thread", "python_type": "bool", "wet_value": False},
    {"setting": "qa_answers"},
    {"setting": "qa_help_text"},
    {"setting": "qa_question"},
    {"setting": "recaptcha_secret_key"},
    {"setting": "recaptcha_site_key", "is_public": True},
    {
        "setting": "signature_length_max",
        "python_type": "int",
        "wet_value": 256,
        "is_public": True,
    },
    {"setting": "subscribe_reply", "dry_value": "watch_email"},
    {"setting": "subscribe_start", "dry_value": "watch_email"},
    {
        "setting": "thread_title_length_max",
        "python_type": "int",
        "wet_value": 90,
        "is_public": True,
    },
    {
        "setting": "thread_title_length_min",
        "python_type": "int",
        "wet_value": 5,
        "is_public": True,
    },
    {"setting": "username_length_min", "python_type": "int", "wet_value": 3},
    {"setting": "username_length_max", "python_type": "int", "wet_value": 14},
    {"setting": "anonymous_username", "dry_value": "Deleted"},
    {"setting": "enable_stop_forum_spam", "python_type": "bool", "wet_value": False},
    {"setting": "stop_forum_spam_confidence", "python_type": "int", "wet_value": 80},
    {"setting": "users_per_page", "python_type": "int", "wet_value": 12},
    {"setting": "users_per_page_orphans", "python_type": "int", "wet_value": 4},
    {"setting": "allow_data_downloads", "python_type": "bool", "wet_value": True},
    {"setting": "data_downloads_expiration", "python_type": "int", "wet_value": 48},
    {"setting": "allow_delete_own_account", "python_type": "bool", "wet_value": False},
    {"setting": "top_posters_ranking_length", "python_type": "int", "wet_value": 30},
    {"setting": "top_posters_ranking_size", "python_type": "int", "wet_value": 50},
    {"setting": "new_inactive_accounts_delete", "python_type": "int", "wet_value": 0},
    {"setting": "ip_storage_time", "python_type": "int", "wet_value": 90},
]

removed_settings = ["forum_branding_display", "forum_branding_text"]


def create_settings(apps, _):
    # This migration builds list of existing settings, and then
    # creates settings not already in the database
    Setting = apps.get_model("misago_conf", "Setting")

    # Update existing settings and add new ones
    existing_settings = list(Setting.objects.values_list("setting", flat=True))
    for setting in default_settings:
        if setting["setting"] in existing_settings:
            continue  # skip already existing setting (migration on existing forum)

        data = setting.copy()
        if "python_type" in data and "wet_value" in data:
            data["dry_value"] = dehydrate_value(
                data["python_type"], data.pop("wet_value")
            )

        Setting.objects.create(**data)

    # Delete deprecated settings
    Setting.objects.filter(setting__in=removed_settings).delete()


class Migration(migrations.Migration):
    dependencies = [
        ("misago_conf", "0003_simplify_models"),
        ("misago_core", "0003_delete_cacheversion"),
        ("misago_threads", "0012_set_dj_partial_indexes"),
        ("misago_users", "0020_set_dj_partial_indexes"),
    ]

    operations = [migrations.RunPython(create_settings)]



# Generated by Django 3.2.15 on 2023-04-06 18:04

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("misago_users", "0023_remove_user_sso_id"),
    ]

    operations = [
        migrations.AddField(
            model_name="user",
            name="notify_new_private_threads_by_followed",
            field=models.PositiveIntegerField(
                choices=[
                    (0, "Don't notify"),
                    (1, "Notify on site only"),
                    (2, "Notify on site and with e-mail"),
                ],
                default=2,
            ),
        ),
        migrations.AddField(
            model_name="user",
            name="notify_new_private_threads_by_other_users",
            field=models.PositiveIntegerField(
                choices=[
                    (0, "Don't notify"),
                    (1, "Notify on site only"),
                    (2, "Notify on site and with e-mail"),
                ],
                default=2,
            ),
        ),
        migrations.AddField(
            model_name="user",
            name="unread_notifications",
            field=models.PositiveIntegerField(default=0),
        ),
        migrations.AddField(
            model_name="user",
            name="watch_new_private_threads_by_followed",
            field=models.PositiveIntegerField(
                choices=[
                    (0, "Don't notify"),
                    (1, "Notify on site only"),
                    (2, "Notify on site and with e-mail"),
                ],
                default=2,
            ),
        ),
        migrations.AddField(
            model_name="user",
            name="watch_new_private_threads_by_other_users",
            field=models.PositiveIntegerField(
                choices=[
                    (0, "Don't notify"),
                    (1, "Notify on site only"),
                    (2, "Notify on site and with e-mail"),
                ],
                default=2,
            ),
        ),
        migrations.AddField(
            model_name="user",
            name="watch_replied_threads",
            field=models.PositiveIntegerField(
                choices=[
                    (0, "Don't notify"),
                    (1, "Notify on site only"),
                    (2, "Notify on site and with e-mail"),
                ],
                default=2,
            ),
        ),
        migrations.AddField(
            model_name="user",
            name="watch_started_threads",
            field=models.PositiveIntegerField(
                choices=[
                    (0, "Don't notify"),
                    (1, "Notify on site only"),
                    (2, "Notify on site and with e-mail"),
                ],
                default=2,
            ),
        ),
    ]



from ..parser import parse


def test_strikethrough_markdown(request_mock, user, snapshot):
    text = "Lorem ~~ipsum~~ dolor met!"
    result = parse(text, request_mock, user)
    assert snapshot == result["parsed_text"]



from datetime import timedelta

from django.utils import timezone

from ...conf.test import override_dynamic_settings
from ...threads.test import reply_thread
from ..threadstracker import make_read_aware


def remove_tracking(thread):
    thread.started_on = timezone.now() - timedelta(days=4)
    thread.save()
    thread.first_post.posted_on = thread.started_on
    thread.first_post.save()


def test_falsy_value_can_be_made_read_aware(request_mock):
    make_read_aware(request_mock, None)
    make_read_aware(request_mock, False)


def test_empty_list_can_be_made_read_aware(request_mock):
    make_read_aware(request_mock, [])


def test_read_thread_is_marked_as_read(request_mock, read_thread):
    make_read_aware(request_mock, read_thread)
    assert read_thread.is_read
    assert not read_thread.is_new


def test_read_thread_with_hidden_post_marked_as_unread(request_mock, read_thread):
    reply_thread(read_thread, is_hidden=True)
    make_read_aware(request_mock, read_thread)
    assert not read_thread.is_read
    assert read_thread.is_new


def test_read_thread_with_invisible_post_marked_as_read(request_mock, read_thread):
    reply_thread(read_thread, is_unapproved=True)
    make_read_aware(request_mock, read_thread)
    assert read_thread.is_read
    assert not read_thread.is_new


def test_read_thread_with_unread_post_marked_as_unread(request_mock, read_thread):
    reply_thread(read_thread)
    make_read_aware(request_mock, read_thread)
    assert not read_thread.is_read
    assert read_thread.is_new


def test_untracked_thread_with_tracked_post_is_marked_as_unread(request_mock, thread):
    remove_tracking(thread)
    reply_thread(thread)
    make_read_aware(request_mock, thread)
    assert not thread.is_read
    assert thread.is_new


def test_tracked_thread_is_marked_as_unread(request_mock, thread):
    make_read_aware(request_mock, thread)
    assert not thread.is_read
    assert thread.is_new


def test_thread_with_post_older_than_user_is_marked_as_read(request_mock, thread, user):
    remove_tracking(thread)
    make_read_aware(request_mock, thread)
    assert thread.is_read
    assert not thread.is_new


@override_dynamic_settings(readtracker_cutoff=3)
def test_non_tracked_thread_is_marked_as_read(request_mock, thread, user):
    user.joined_on = timezone.now() - timedelta(days=10)
    user.save()

    remove_tracking(thread)
    make_read_aware(request_mock, thread)
    assert thread.is_read
    assert not thread.is_new


def test_read_thread_with_new_event_is_marked_as_unread(request_mock, read_thread):
    reply_thread(read_thread, is_event=True)
    make_read_aware(request_mock, read_thread)
    assert not read_thread.is_read
    assert read_thread.is_new


def test_read_thread_with_hidden_event_is_marked_as_read(request_mock, read_thread):
    reply_thread(read_thread, is_hidden=True, is_event=True)
    make_read_aware(request_mock, read_thread)
    assert read_thread.is_read
    assert not read_thread.is_new


def test_read_thread_with_hidden_event_visible_to_user_is_marked_as_unread(
    request_mock, read_thread, default_category
):
    request_mock.user_acl["categories"][default_category.id]["can_hide_events"] = 1
    reply_thread(read_thread, is_hidden=True, is_event=True)
    make_read_aware(request_mock, read_thread)
    assert not read_thread.is_read
    assert read_thread.is_new


def test_tracked_thread_is_marked_as_read_for_anonymous_user(
    anonymous_request_mock, thread
):
    make_read_aware(anonymous_request_mock, thread)
    assert thread.is_read
    assert not thread.is_new


@override_dynamic_settings(readtracker_cutoff=3)
def test_non_tracked_thread_is_marked_as_read_for_anonymous_user(
    anonymous_request_mock, thread
):
    remove_tracking(thread)
    make_read_aware(anonymous_request_mock, thread)
    assert thread.is_read
    assert not thread.is_new



from django.shortcuts import get_object_or_404, redirect, render
from django.urls import reverse
from django.views import View

from ...core.utils import format_plaintext_for_html
from ..models import Rank
from ..pages import users_list
from ..permissions import allow_browse_users_list
from ..viewmodels import ActivePosters, RankUsers


class ListView(View):
    def get(self, request, *args, **kwargs):
        allow_browse_users_list(request.user_acl)

        context_data = self.get_context_data(request, *args, **kwargs)

        sections = users_list.get_sections(request)

        context_data["pages"] = sections

        request.frontend_context["USERS_LISTS"] = []
        for page in sections:
            page["reversed_link"] = reverse(page["link"])
            request.frontend_context["USERS_LISTS"].append(
                {"name": str(page["name"]), "component": page["component"]}
            )

        active_rank = context_data.get("rank")
        for rank in Rank.objects.filter(is_tab=True).order_by("order"):
            context_data["pages"].append(
                {
                    "name": str(rank),
                    "reversed_link": reverse(
                        "misago:users-rank", kwargs={"slug": rank.slug}
                    ),
                    "is_active": active_rank.pk == rank.pk if active_rank else None,
                }
            )

            if rank.description:
                description = {
                    "plain": rank.description,
                    "html": format_plaintext_for_html(rank.description),
                }
            else:
                description = None

            request.frontend_context["USERS_LISTS"].append(
                {
                    "id": rank.pk,
                    "name": str(rank),
                    "slug": rank.slug,
                    "css_class": rank.css_class,
                    "description": description,
                    "component": "rank",
                }
            )

        active_section = list(filter(lambda x: x["is_active"], sections))[0]
        context_data["active_section"] = active_section

        return render(request, self.template_name, context_data)

    def get_context_data(self, request, *args, **kwargs):
        return {}


def landing(request):
    allow_browse_users_list(request.user_acl)
    return redirect(users_list.get_default_link())


class ActivePostersView(ListView):
    template_name = "misago/userslists/active_posters.html"

    def get_context_data(self, request, *args, **kwargs):
        model = ActivePosters(request)

        request.frontend_context["USERS"] = model.get_frontend_context()

        return model.get_template_context()


class RankUsersView(ListView):
    template_name = "misago/userslists/rank.html"

    def get_context_data(self, request, slug, page=0):
        rank = get_object_or_404(Rank.objects.filter(is_tab=True), slug=slug)
        users = RankUsers(request, rank, page)

        request.frontend_context["USERS"] = users.get_frontend_context()

        context = {"rank": rank}
        context.update(users.get_template_context())

        return context



import re
from xml.etree.ElementTree import SubElement

import markdown
from django.utils.crypto import get_random_string
from markdown.blockprocessors import BlockProcessor
from markdown.preprocessors import Preprocessor

SPOILER_START = get_random_string(32)
SPOILER_END = get_random_string(32)


class SpoilerExtension(markdown.Extension):
    def extendMarkdown(self, md):
        md.registerExtension(self)

        md.preprocessors.register(SpoilerPreprocessor(md), "misago_bbcode_spoiler", 200)
        md.parser.blockprocessors.register(
            SpoilerBlockProcessor(md.parser), "misago_bbcode_spoiler", 85
        )


class SpoilerPreprocessor(Preprocessor):
    SPOILER_BLOCK_RE = re.compile(
        r"""
\[spoiler\](?P<text>.*?)\[/spoiler\]
""".strip(),
        re.IGNORECASE | re.MULTILINE | re.DOTALL,
    )

    def run(self, lines):
        text = "\n".join(lines)
        while self.SPOILER_BLOCK_RE.search(text):
            text = self.SPOILER_BLOCK_RE.sub(self.replace, text)
        return text.split("\n")

    def replace(self, matchobj):
        text = matchobj.group("text")
        return "\n\n%s\n\n%s\n\n%s\n\n" % (SPOILER_START, text, SPOILER_END)


class SpoilerBlockProcessor(BlockProcessor):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._spoiler = 0
        self._children = []

    def test(self, parent, block):
        return block.strip().startswith(SPOILER_START) or self._spoiler

    def run(self, parent, blocks):
        block = blocks.pop(0)
        if block.strip().startswith(SPOILER_START):
            self._spoiler += 1

        self._children.append(block)

        if block.strip() == SPOILER_END:
            self._spoiler -= 1

        if not self._spoiler:
            children, self._children = self._children[1:-1], []

            aside = SubElement(parent, "aside")
            aside.set("class", "spoiler-block")

            blockquote = SubElement(aside, "blockquote")
            blockquote.set("class", "spoiler-body")
            blockquote.set("data-block", "spoiler")

            overlay = SubElement(aside, "div")
            overlay.set("class", "spoiler-overlay")
            overlay.set("data-noquote", "1")

            reveal_btn = SubElement(overlay, "button")
            reveal_btn.set("class", "spoiler-reveal")
            reveal_btn.set("type", "button")

            self.parser.parseBlocks(blockquote, children)



from django import forms
from django.utils.translation import pgettext_lazy

from ..acl import algebra
from ..acl.models import Role
from ..admin.forms import YesNoSwitch


class PermissionsForm(forms.Form):
    legend = pgettext_lazy("search permission", "Search")

    can_search = YesNoSwitch(
        label=pgettext_lazy("search permission", "Can search site"), initial=1
    )


def change_permissions_form(role):
    if isinstance(role, Role):
        return PermissionsForm


def build_acl(acl, roles, key_name):
    new_acl = {"can_search": 0}
    new_acl.update(acl)

    return algebra.sum_acls(
        new_acl, roles=roles, key=key_name, can_search=algebra.greater
    )



from urllib.parse import urlencode
from typing import Any

import requests
from django.urls import reverse
from django.utils.crypto import get_random_string
from requests.exceptions import RequestException

from . import exceptions

SESSION_STATE = "oauth2_state"
STATE_LENGTH = 40
REQUESTS_TIMEOUT = 30


def create_login_url(request):
    state = get_random_string(STATE_LENGTH)
    request.session[SESSION_STATE] = state

    quote = {
        "response_type": "code",
        "client_id": request.settings.oauth2_client_id,
        "redirect_uri": get_redirect_uri(request),
        "scope": request.settings.oauth2_scopes,
        "state": state,
    }

    return "%s?%s" % (request.settings.oauth2_login_url, urlencode(quote))


def get_code_grant(request):
    session_state = request.session.pop(SESSION_STATE, None)

    if request.GET.get("error") == "access_denied":
        raise exceptions.OAuth2AccessDeniedError()

    if not session_state:
        raise exceptions.OAuth2StateNotSetError()

    provider_state = request.GET.get("state")
    if not provider_state:
        raise exceptions.OAuth2StateNotProvidedError()
    if provider_state != session_state:
        raise exceptions.OAuth2StateMismatchError()

    code_grant = request.GET.get("code")
    if not code_grant:
        raise exceptions.OAuth2CodeNotProvidedError()

    return code_grant


def get_access_token(request, code_grant):
    token_url = request.settings.oauth2_token_url
    data = {
        "grant_type": "authorization_code",
        "client_id": request.settings.oauth2_client_id,
        "client_secret": request.settings.oauth2_client_secret,
        "redirect_uri": get_redirect_uri(request),
        "code": code_grant,
    }
    headers = get_headers_dict(request.settings.oauth2_token_extra_headers)

    try:
        r = requests.post(
            token_url,
            data=data,
            headers=headers,
            timeout=REQUESTS_TIMEOUT,
        )
    except RequestException:
        raise exceptions.OAuth2AccessTokenRequestError()

    if r.status_code != 200:
        raise exceptions.OAuth2AccessTokenResponseError()

    try:
        response_json = r.json()
        if not isinstance(response_json, dict):
            raise TypeError()
    except (ValueError, TypeError):
        raise exceptions.OAuth2AccessTokenJSONError()

    access_token = get_value_from_json(
        request.settings.oauth2_json_token_path,
        response_json,
    )

    if not access_token:
        raise exceptions.OAuth2AccessTokenNotProvidedError()

    return access_token


JSON_MAPPING = {
    "id": "oauth2_json_id_path",
    "name": "oauth2_json_name_path",
    "email": "oauth2_json_email_path",
    "avatar": "oauth2_json_avatar_path",
}


def get_user_data(request, access_token):
    headers = get_headers_dict(request.settings.oauth2_user_extra_headers)
    user_url = request.settings.oauth2_user_url

    if request.settings.oauth2_user_token_location == "QUERY":
        user_url += "&" if "?" in user_url else "?"
        user_url += urlencode({request.settings.oauth2_user_token_name: access_token})
    elif request.settings.oauth2_user_token_location == "HEADER_BEARER":
        headers[request.settings.oauth2_user_token_name] = f"Bearer {access_token}"
    else:
        headers[request.settings.oauth2_user_token_name] = access_token

    try:
        if request.settings.oauth2_user_method == "GET":
            r = requests.get(user_url, headers=headers, timeout=REQUESTS_TIMEOUT)
        else:
            r = requests.post(user_url, headers=headers, timeout=REQUESTS_TIMEOUT)
    except RequestException:
        raise exceptions.OAuth2UserDataRequestError()

    if r.status_code != 200:
        raise exceptions.OAuth2UserDataResponseError()

    try:
        response_json = r.json()
        if not isinstance(response_json, dict):
            raise TypeError()
    except (ValueError, TypeError):
        raise exceptions.OAuth2UserDataJSONError()

    user_data = {
        key: get_value_from_json(getattr(request.settings, setting), response_json)
        for key, setting in JSON_MAPPING.items()
    }

    return user_data, response_json


def get_redirect_uri(request):
    return request.build_absolute_uri(reverse("misago:oauth2-complete"))


def get_headers_dict(headers_str):
    headers = {}
    if not headers_str:
        return headers

    for header in headers_str.splitlines():
        header = header.strip()
        if ":" not in header:
            continue

        header_name, header_value = [part.strip() for part in header.split(":", 1)]
        if header_name and header_value:
            headers[header_name] = header_value

    return headers


def get_value_from_json(path, json):
    if not path:
        return None

    if "." not in path:
        return clear_json_value(json.get(path))

    data = json
    for path_part in path.split("."):
        if not isinstance(data, dict):
            return None

        data = data.get(path_part)
        if data is None:
            return None

    return clear_json_value(data)


def clear_json_value(value: Any) -> str | None:
    if isinstance(value, str):
        return value.strip() or None

    if isinstance(value, int) and value is not True and value is not False:
        return str(value)

    return None



from django.test import TestCase
from django.urls import reverse

from ...conf.test import override_dynamic_settings

test_qa_question = "Do you like pies?"
test_qa_help_text = 'Type in "yes".'


class AuthenticateApiTests(TestCase):
    def setUp(self):
        self.api_link = reverse("misago:api:captcha-question")

    @override_dynamic_settings(qa_question="")
    def test_api_no_qa_is_set(self):
        """qa api returns 404 if no QA question is set"""
        response = self.client.get(self.api_link)
        self.assertEqual(response.status_code, 404)

    @override_dynamic_settings(
        qa_question=test_qa_question, qa_help_text=test_qa_help_text
    )
    def test_api_get_question(self):
        """qa api returns valid QA question"""
        response = self.client.get(self.api_link)
        self.assertEqual(response.status_code, 200)

        response_json = response.json()
        self.assertEqual(response_json["question"], test_qa_question)
        self.assertEqual(response_json["help_text"], test_qa_help_text)



# Generated by Django 4.2.7 on 2023-12-05 19:22

import django.contrib.postgres.indexes
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        (
            "misago_threads",
            "0013_rename_pollvote_poll_voter_name_misago_thre_poll_id_c3e8fe_idx_and_more",
        ),
    ]

    operations = [
        migrations.AddField(
            model_name="attachment",
            name="plugin_data",
            field=models.JSONField(default=dict),
        ),
        migrations.AddField(
            model_name="attachmenttype",
            name="plugin_data",
            field=models.JSONField(default=dict),
        ),
        migrations.AddField(
            model_name="poll",
            name="plugin_data",
            field=models.JSONField(default=dict),
        ),
        migrations.AddField(
            model_name="post",
            name="plugin_data",
            field=models.JSONField(default=dict),
        ),
        migrations.AddField(
            model_name="thread",
            name="plugin_data",
            field=models.JSONField(default=dict),
        ),
        migrations.AddIndex(
            model_name="attachment",
            index=django.contrib.postgres.indexes.GinIndex(
                fields=["plugin_data"], name="misago_thre_plugin__46d25a_gin"
            ),
        ),
        migrations.AddIndex(
            model_name="attachmenttype",
            index=django.contrib.postgres.indexes.GinIndex(
                fields=["plugin_data"], name="misago_thre_plugin__236db5_gin"
            ),
        ),
        migrations.AddIndex(
            model_name="poll",
            index=django.contrib.postgres.indexes.GinIndex(
                fields=["plugin_data"], name="misago_thre_plugin__5e1be7_gin"
            ),
        ),
        migrations.AddIndex(
            model_name="post",
            index=django.contrib.postgres.indexes.GinIndex(
                fields=["plugin_data"], name="misago_thre_plugin__91da06_gin"
            ),
        ),
        migrations.AddIndex(
            model_name="thread",
            index=django.contrib.postgres.indexes.GinIndex(
                fields=["plugin_data"], name="misago_thre_plugin__e6e836_gin"
            ),
        ),
    ]



from datetime import timedelta

from django.contrib.auth import get_user_model
from django.db.models import Count
from django.utils import timezone

from ..categories.models import Category
from ..conf.shortcuts import get_dynamic_settings
from .models import ActivityRanking

User = get_user_model()


def get_active_posters_ranking():
    users = []

    queryset = ActivityRanking.objects.select_related("user", "user__rank")
    for ranking in queryset.order_by("-score"):
        ranking.user.score = ranking.score
        users.append(ranking.user)

    return {"users": users, "users_count": len(users)}


def build_active_posters_ranking() -> list[ActivityRanking]:
    settings = get_dynamic_settings()
    tracked_period = settings.top_posters_ranking_length
    tracked_since = timezone.now() - timedelta(days=tracked_period)

    ActivityRanking.objects.all().delete()

    ranked_categories = []
    for category in Category.objects.all_categories():
        ranked_categories.append(category.pk)

    queryset = (
        User.objects.filter(
            is_active=True,
            post__is_event=False,
            post__posted_on__gte=tracked_since,
            post__category__in=ranked_categories,
        )
        .annotate(score=Count("post"))
        .filter(score__gt=0)
        .order_by("-score")
    )[: settings.top_posters_ranking_size]

    new_ranking = []
    for ranking in queryset.iterator():
        new_ranking.append(ActivityRanking(user=ranking, score=ranking.score))

    ActivityRanking.objects.bulk_create(new_ranking)

    return new_ranking



from ..dynamicsettings import DynamicSettings
from ..test import override_dynamic_settings


def test_dynamic_setting_can_be_overridden_using_context_manager(dynamic_settings):
    assert dynamic_settings.forum_name == "Misago"
    with override_dynamic_settings(forum_name="Overrided"):
        assert dynamic_settings.forum_name == "Overrided"
    assert dynamic_settings.forum_name == "Misago"


def test_dynamic_setting_can_be_overridden_using_decorator(dynamic_settings):
    @override_dynamic_settings(forum_name="Overrided")
    def decorated_function(settings):
        return settings.forum_name

    assert dynamic_settings.forum_name == "Misago"
    assert decorated_function(dynamic_settings) == "Overrided"
    assert dynamic_settings.forum_name == "Misago"


def test_lazy_dynamic_setting_can_be_overridden_using_context_manager(
    cache_versions, lazy_setting
):
    settings = DynamicSettings(cache_versions)
    assert settings.get_lazy_setting_value("lazy_setting") == "Hello"
    with override_dynamic_settings(lazy_setting="Overrided"):
        assert settings.get_lazy_setting_value("lazy_setting") == "Overrided"
    assert settings.get_lazy_setting_value("lazy_setting") == "Hello"


def test_lazy_dynamic_setting_can_be_overridden_using_decorator(
    cache_versions, lazy_setting
):
    @override_dynamic_settings(lazy_setting="Overrided")
    def decorated_function(settings):
        return settings.get_lazy_setting_value("lazy_setting")

    settings = DynamicSettings(cache_versions)
    assert settings.get_lazy_setting_value("lazy_setting") == "Hello"
    assert decorated_function(settings) == "Overrided"
    assert settings.get_lazy_setting_value("lazy_setting") == "Hello"



from django.contrib.auth import get_user_model

from ...conf.test import override_dynamic_settings
from ...notifications.threads import ThreadNotifications
from ..setupnewuser import setup_new_user

User = get_user_model()

AVATAR_URL = "https://dummyimage.com/600/500"


def user(db):
    return User.objects.create_user("User", "user@example.com")


def test_default_avatar_is_set_for_user(dynamic_settings, user):
    setup_new_user(dynamic_settings, user)
    assert user.avatars
    assert user.avatar_set.exists()


def test_avatar_from_url_is_set_for_user(dynamic_settings, user):
    setup_new_user(dynamic_settings, user, avatar_url=AVATAR_URL)
    assert user.avatars
    assert user.avatar_set.exists()


def test_default_started_threads_subscription_option_is_set_for_user(
    dynamic_settings, user
):
    with override_dynamic_settings(subscribe_start="no"):
        setup_new_user(dynamic_settings, user, avatar_url=AVATAR_URL)
        assert user.subscribe_to_started_threads == User.SUBSCRIPTION_NONE

    with override_dynamic_settings(subscribe_start="watch"):
        setup_new_user(dynamic_settings, user, avatar_url=AVATAR_URL)
        assert user.subscribe_to_started_threads == User.SUBSCRIPTION_NOTIFY

    with override_dynamic_settings(subscribe_start="watch_email"):
        setup_new_user(dynamic_settings, user, avatar_url=AVATAR_URL)
        assert user.subscribe_to_started_threads == User.SUBSCRIPTION_ALL


def test_default_replied_threads_subscription_option_is_set_for_user(
    dynamic_settings, user
):
    with override_dynamic_settings(subscribe_reply="no"):
        setup_new_user(dynamic_settings, user, avatar_url=AVATAR_URL)
        assert user.subscribe_to_replied_threads == User.SUBSCRIPTION_NONE

    with override_dynamic_settings(subscribe_reply="watch"):
        setup_new_user(dynamic_settings, user, avatar_url=AVATAR_URL)
        assert user.subscribe_to_replied_threads == User.SUBSCRIPTION_NOTIFY

    with override_dynamic_settings(subscribe_reply="watch_email"):
        setup_new_user(dynamic_settings, user, avatar_url=AVATAR_URL)
        assert user.subscribe_to_replied_threads == User.SUBSCRIPTION_ALL


def test_default_watch_started_threads_option_is_set_for_user(dynamic_settings, user):
    with override_dynamic_settings(watch_started_threads=ThreadNotifications.NONE):
        setup_new_user(dynamic_settings, user, avatar_url=AVATAR_URL)
        assert user.watch_started_threads == ThreadNotifications.NONE

    with override_dynamic_settings(watch_started_threads=ThreadNotifications.SITE_ONLY):
        setup_new_user(dynamic_settings, user, avatar_url=AVATAR_URL)
        assert user.watch_started_threads == ThreadNotifications.SITE_ONLY

    with override_dynamic_settings(
        watch_started_threads=ThreadNotifications.SITE_AND_EMAIL
    ):
        setup_new_user(dynamic_settings, user, avatar_url=AVATAR_URL)
        assert user.watch_started_threads == ThreadNotifications.SITE_AND_EMAIL


def test_default_watch_replied_threads_option_is_set_for_user(dynamic_settings, user):
    with override_dynamic_settings(watch_replied_threads=ThreadNotifications.NONE):
        setup_new_user(dynamic_settings, user, avatar_url=AVATAR_URL)
        assert user.watch_replied_threads == ThreadNotifications.NONE

    with override_dynamic_settings(watch_replied_threads=ThreadNotifications.SITE_ONLY):
        setup_new_user(dynamic_settings, user, avatar_url=AVATAR_URL)
        assert user.watch_replied_threads == ThreadNotifications.SITE_ONLY

    with override_dynamic_settings(
        watch_replied_threads=ThreadNotifications.SITE_AND_EMAIL
    ):
        setup_new_user(dynamic_settings, user, avatar_url=AVATAR_URL)
        assert user.watch_replied_threads == ThreadNotifications.SITE_AND_EMAIL


def test_default_watch_new_private_threads_by_followed_option_is_set_for_user(
    dynamic_settings, user
):
    with override_dynamic_settings(
        watch_new_private_threads_by_followed=ThreadNotifications.NONE
    ):
        setup_new_user(dynamic_settings, user, avatar_url=AVATAR_URL)
        assert user.watch_new_private_threads_by_followed == ThreadNotifications.NONE

    with override_dynamic_settings(
        watch_new_private_threads_by_followed=ThreadNotifications.SITE_ONLY
    ):
        setup_new_user(dynamic_settings, user, avatar_url=AVATAR_URL)
        assert (
            user.watch_new_private_threads_by_followed == ThreadNotifications.SITE_ONLY
        )

    with override_dynamic_settings(
        watch_new_private_threads_by_followed=ThreadNotifications.SITE_AND_EMAIL
    ):
        setup_new_user(dynamic_settings, user, avatar_url=AVATAR_URL)
        assert (
            user.watch_new_private_threads_by_followed
            == ThreadNotifications.SITE_AND_EMAIL
        )


def test_default_watch_new_private_threads_by_other_users_option_is_set_for_user(
    dynamic_settings, user
):
    with override_dynamic_settings(
        watch_new_private_threads_by_other_users=ThreadNotifications.NONE
    ):
        setup_new_user(dynamic_settings, user, avatar_url=AVATAR_URL)
        assert user.watch_new_private_threads_by_other_users == ThreadNotifications.NONE

    with override_dynamic_settings(
        watch_new_private_threads_by_other_users=ThreadNotifications.SITE_ONLY
    ):
        setup_new_user(dynamic_settings, user, avatar_url=AVATAR_URL)
        assert (
            user.watch_new_private_threads_by_other_users
            == ThreadNotifications.SITE_ONLY
        )

    with override_dynamic_settings(
        watch_new_private_threads_by_other_users=ThreadNotifications.SITE_AND_EMAIL
    ):
        setup_new_user(dynamic_settings, user, avatar_url=AVATAR_URL)
        assert (
            user.watch_new_private_threads_by_other_users
            == ThreadNotifications.SITE_AND_EMAIL
        )


def test_default_notify_new_private_threads_by_followed_option_is_set_for_user(
    dynamic_settings, user
):
    with override_dynamic_settings(
        notify_new_private_threads_by_followed=ThreadNotifications.NONE
    ):
        setup_new_user(dynamic_settings, user, avatar_url=AVATAR_URL)
        assert user.notify_new_private_threads_by_followed == ThreadNotifications.NONE

    with override_dynamic_settings(
        notify_new_private_threads_by_followed=ThreadNotifications.SITE_ONLY
    ):
        setup_new_user(dynamic_settings, user, avatar_url=AVATAR_URL)
        assert (
            user.notify_new_private_threads_by_followed == ThreadNotifications.SITE_ONLY
        )

    with override_dynamic_settings(
        notify_new_private_threads_by_followed=ThreadNotifications.SITE_AND_EMAIL
    ):
        setup_new_user(dynamic_settings, user, avatar_url=AVATAR_URL)
        assert (
            user.notify_new_private_threads_by_followed
            == ThreadNotifications.SITE_AND_EMAIL
        )


def test_default_notify_new_private_threads_by_other_users_option_is_set_for_user(
    dynamic_settings, user
):
    with override_dynamic_settings(
        notify_new_private_threads_by_other_users=ThreadNotifications.NONE
    ):
        setup_new_user(dynamic_settings, user, avatar_url=AVATAR_URL)
        assert (
            user.notify_new_private_threads_by_other_users == ThreadNotifications.NONE
        )

    with override_dynamic_settings(
        notify_new_private_threads_by_other_users=ThreadNotifications.SITE_ONLY
    ):
        setup_new_user(dynamic_settings, user, avatar_url=AVATAR_URL)
        assert (
            user.notify_new_private_threads_by_other_users
            == ThreadNotifications.SITE_ONLY
        )

    with override_dynamic_settings(
        notify_new_private_threads_by_other_users=ThreadNotifications.SITE_AND_EMAIL
    ):
        setup_new_user(dynamic_settings, user, avatar_url=AVATAR_URL)
        assert (
            user.notify_new_private_threads_by_other_users
            == ThreadNotifications.SITE_AND_EMAIL
        )


def test_if_user_ip_is_available_audit_trail_is_created_for_user(dynamic_settings):
    user = User.objects.create_user(
        "User", "user@example.com", joined_from_ip="0.0.0.0"
    )
    setup_new_user(dynamic_settings, user)
    assert user.audittrail_set.count() == 1


def test_if_user_ip_is_not_available_audit_trail_is_not_created(dynamic_settings, user):
    setup_new_user(dynamic_settings, user)
    assert user.audittrail_set.exists() is False



from rest_framework import serializers

from . import PostingEndpoint, PostingMiddleware
from ... import moderation
from ...models import Thread


class PinMiddleware(PostingMiddleware):
    def use_this_middleware(self):
        return self.mode == PostingEndpoint.START

    def get_serializer(self):
        return PinSerializer(data=self.request.data)

    def post_save(self, serializer):
        allowed_pin = self.thread.category.acl["can_pin_threads"]
        if allowed_pin > 0:
            pin = serializer.validated_data["pin"]

            if pin <= allowed_pin:
                if pin == Thread.WEIGHT_GLOBAL:
                    moderation.pin_thread_globally(self.request, self.thread)
                elif pin == Thread.WEIGHT_PINNED:
                    moderation.pin_thread_locally(self.request, self.thread)


class PinSerializer(serializers.Serializer):
    pin = serializers.IntegerField(required=False, default=0)



from django.core.cache import cache
from django.db import models
from django.urls import reverse
from django.utils import timezone
from django.utils.translation import pgettext_lazy

from ..conf import settings

CACHE_KEY = "misago_agreements"


class AgreementManager(models.Manager):
    def invalidate_cache(self):
        cache.delete(CACHE_KEY)

    def get_agreements(self):
        agreements = self.get_agreements_from_cache()
        if agreements == "nada":
            agreements = self.get_agreements_from_db()
            cache.set(CACHE_KEY, agreements)
        return agreements

    def get_agreements_from_cache(self):
        return cache.get(CACHE_KEY, "nada")

    def get_agreements_from_db(self):
        agreements = {}
        for agreement in Agreement.objects.filter(is_active=True):
            agreements[agreement.type] = {
                "id": agreement.id,
                "title": agreement.get_final_title(),
                "link": agreement.link,
                "text": bool(agreement.text),
            }
        return agreements


class Agreement(models.Model):
    TYPE_TOS = "terms_of_service"
    TYPE_PRIVACY = "privacy_policy"
    TYPE_CHOICES = [
        (TYPE_TOS, pgettext_lazy("agreement type choice", "Terms of service")),
        (TYPE_PRIVACY, pgettext_lazy("agreement type choice", "Privacy policy")),
    ]

    type = models.CharField(
        max_length=20, default=TYPE_TOS, choices=TYPE_CHOICES, db_index=True
    )
    title = models.CharField(max_length=255, null=True, blank=True)
    link = models.URLField(max_length=255, null=True, blank=True)
    text = models.TextField(null=True, blank=True)
    is_active = models.BooleanField(default=False)
    created_on = models.DateTimeField(default=timezone.now)
    created_by = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.SET_NULL,
        blank=True,
        null=True,
        related_name="+",
    )
    created_by_name = models.CharField(max_length=255, null=True, blank=True)
    last_modified_on = models.DateTimeField(null=True, blank=True)
    last_modified_by = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.SET_NULL,
        blank=True,
        null=True,
        related_name="+",
    )
    last_modified_by_name = models.CharField(max_length=255, null=True, blank=True)

    objects = AgreementManager()

    def get_absolute_url(self):
        if not self.is_active:
            return None

        if self.link:
            return self.link

        if self.type == self.TYPE_TOS:
            return reverse("misago:terms-of-service")

        return reverse("misago:privacy-policy")

    def get_final_title(self):
        return self.title or self.get_type_display()

    def set_created_by(self, user):
        self.created_by = user
        self.created_by_name = user.username

    def set_last_modified_by(self, user):
        self.last_modified_by = user
        self.last_modified_by_name = user.username


class UserAgreement(models.Model):
    user = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.CASCADE)
    agreement = models.ForeignKey(
        Agreement, related_name="accepted_by", on_delete=models.CASCADE
    )
    accepted_on = models.DateTimeField(default=timezone.now)

    class Meta:
        ordering = ["-pk"]



from django.urls import reverse

from ...notifications.models import WatchedThread
from ...notifications.threads import ThreadNotifications


def test_started_thread_is_watched_by_user_with_option_enabled(
    user, user_client, default_category
):
    user.watch_started_threads = ThreadNotifications.SITE_ONLY
    user.save()

    response = user_client.post(
        reverse("misago:api:thread-list"),
        data={
            "category": default_category.pk,
            "title": "Test thread",
            "post": "Lorem ipsum dolor met",
        },
    )

    assert response.status_code == 200

    data = response.json()

    WatchedThread.objects.get(
        user=user,
        category=default_category,
        thread_id=data["id"],
        send_emails=False,
    )


def test_started_thread_is_not_watched_by_user_with_option_disabled(
    user, user_client, default_category
):
    user.watch_started_threads = ThreadNotifications.NONE
    user.save()

    response = user_client.post(
        reverse("misago:api:thread-list"),
        data={
            "category": default_category.pk,
            "title": "Test thread",
            "post": "Lorem ipsum dolor met",
        },
    )

    assert response.status_code == 200
    assert not WatchedThread.objects.exists()


def test_started_private_thread_is_watched_by_user_with_option_enabled(
    notify_on_new_private_thread_mock,
    user,
    user_client,
    other_user,
    private_threads_category,
):
    user.watch_started_threads = ThreadNotifications.SITE_ONLY
    user.save()

    response = user_client.post(
        reverse("misago:api:private-thread-list"),
        data={
            "to": [other_user.username],
            "title": "Test thread",
            "post": "Lorem ipsum dolor met",
        },
    )

    assert response.status_code == 200

    data = response.json()

    WatchedThread.objects.get(
        user=user,
        category=private_threads_category,
        thread_id=data["id"],
        send_emails=False,
    )


def test_started_private_thread_is_not_watched_by_user_with_option_disabled(
    notify_on_new_private_thread_mock, user, user_client, other_user
):
    user.watch_started_threads = ThreadNotifications.NONE
    user.save()

    response = user_client.post(
        reverse("misago:api:private-thread-list"),
        data={
            "to": [other_user.username],
            "title": "Test thread",
            "post": "Lorem ipsum dolor met",
        },
    )

    assert response.status_code == 200
    assert not WatchedThread.objects.exists()



from django.urls import reverse

from ...conf.test import override_dynamic_settings
from ...test import assert_contains


@override_dynamic_settings(google_site_verification=None)
def test_verification_page_returns_404_if_verification_token_is_not_set(db, client):
    response = client.get("/googlet0k3n.html")
    assert response.status_code == 404


@override_dynamic_settings(google_site_verification="t0k3n")
def test_verification_page_returns_404_if_verification_token_is_invalid(db, client):
    response = client.get("/googleinv4l1d.html")
    assert response.status_code == 404


@override_dynamic_settings(google_site_verification="t0k3n")
def test_verification_page_returns_google_verification_token(db, client):
    response = client.get("/googlet0k3n.html")
    assert response.status_code == 200
    assert_contains(response, f"google-site-verification: googlet0k3n.html")



from django.core.management.base import BaseCommand

from ...versions import invalidate_all_caches


class Command(BaseCommand):
    help = "Invalidates versioned caches"

    def handle(self, *args, **options):
        invalidate_all_caches()
        self.stdout.write("Invalidated all versioned caches.")



from django.core.management.base import BaseCommand
from django.utils import timezone

from ....cache.versions import get_cache_versions
from ....users import BANS_CACHE
from ...models import Ban, BanCache


class Command(BaseCommand):
    help = (
        "Runs maintenance on Misago bans system, "
        "invalidating expired bans and pruning caches."
    )

    def handle(self, *args, **options):
        self.handle_expired_bans()
        self.handle_bans_caches()

    def handle_expired_bans(self):
        queryset = Ban.objects.filter(is_checked=True)
        queryset = queryset.filter(expires_on__lt=timezone.now())

        expired_count = queryset.update(is_checked=False)
        self.stdout.write("Bans invalidated: %s" % expired_count)

    def handle_bans_caches(self):
        queryset = BanCache.objects.filter(expires_on__lt=timezone.now())

        expired_count = queryset.count()
        queryset.delete()

        cache_versions = get_cache_versions()
        queryset = BanCache.objects.exclude(cache_version=cache_versions[BANS_CACHE])

        expired_count += queryset.count()
        queryset.delete()

        self.stdout.write("Ban caches emptied: %s" % expired_count)



from django.db import IntegrityError
from django.db.transaction import TransactionManagementError


def retry_on_db_error(f):
    def wrapper(*args, **kwargs):
        try:
            return f(*args, **kwargs)
        except (IntegrityError, TransactionManagementError):
            return wrapper(*args, **kwargs)

    return wrapper



class SearchProvider:
    def __init__(self, request):
        self.request = request

    def allow_search(self):
        pass

    def search(self, query, page=1):
        raise NotImplementedError(
            "%s has to define search(query, page=1) method" % self.__class__.__name__
        )



# Generated by Django 3.2.15 on 2023-04-04 18:29

from django.db import migrations

from ..hydrators import dehydrate_value

setting = {
    "setting": "show_admin_panel_link_in_ui",
    "python_type": "bool",
    "wet_value": True,
    "is_public": False,
}


def create_settings(apps, _):
    Setting = apps.get_model("misago_conf", "Setting")

    Setting.objects.create(
        setting=setting["setting"],
        python_type=setting["python_type"],
        dry_value=dehydrate_value(setting["python_type"], setting["wet_value"]),
        is_public=setting["is_public"],
    )


def delete_settings(apps, _):
    Setting = apps.get_model("misago_conf", "Setting")

    Setting.objects.filter(setting=setting["setting"]).delete()


class Migration(migrations.Migration):
    dependencies = [
        ("misago_conf", "0009_delete_oauth2_access_token_method"),
    ]

    operations = [migrations.RunPython(create_settings, delete_settings)]



from django.core.exceptions import PermissionDenied
from django.db import transaction
from django.http import Http404
from rest_framework.response import Response

ALLOWED_OPS = ("add", "remove", "replace")


class InvalidAction(Exception):
    pass


class ApiPatch:
    def __init__(self):
        self._actions = []

    def add(self, path, handler):
        self._actions.append({"op": "add", "path": path, "handler": handler})

    def remove(self, path, handler):
        self._actions.append({"op": "remove", "path": path, "handler": handler})

    def replace(self, path, handler):
        self._actions.append({"op": "replace", "path": path, "handler": handler})

    def dispatch(self, request, target):
        if not isinstance(request.data, list):
            return Response(
                {"detail": "PATCH request should be list of operations"}, status=400
            )

        detail = []
        is_errored = False

        patch = {"id": target.pk}
        for action in request.data:
            try:
                self.validate_action(action)
                self.dispatch_action(patch, request, target, action)
                detail.append("ok")
            except Http404:
                is_errored = True
                detail.append("NOT FOUND")
                break
            except (InvalidAction, PermissionDenied) as e:
                is_errored = True
                detail.append(e.args[0])
                break

        patch["detail"] = detail
        if is_errored:
            return Response(patch, status=400)
        return Response(patch)

    def dispatch_bulk(self, request, targets):
        is_errored = False
        result = []

        for target in targets:
            detail = []

            patch = {"id": target.pk}
            for action in request.data["ops"]:
                try:
                    self.validate_action(action)
                    self.dispatch_action(patch, request, target, action)
                except Http404:
                    is_errored = True
                    detail.append("NOT FOUND")
                    break
                except (InvalidAction, PermissionDenied) as e:
                    is_errored = True
                    detail.append(e.args[0])
                    break
            if detail:
                patch["detail"] = detail
            result.append(patch)

        if is_errored:
            return Response(result, status=400)
        return Response(result)

    def validate_action(self, action):
        if not action.get("op"):
            raise InvalidAction("undefined op")

        if action.get("op") not in ALLOWED_OPS:
            raise InvalidAction('"%s" op is unsupported' % action.get("op"))

        if not action.get("path"):
            raise InvalidAction('"%s" op has to specify path' % action.get("op"))

        if "value" not in action:
            raise InvalidAction('"%s" op has to specify value' % action.get("op"))

    def dispatch_action(self, patch, request, target, action):
        for handler in self._actions:
            if action["op"] == handler["op"] and action["path"] == handler["path"]:
                with transaction.atomic():
                    patch.update(handler["handler"](request, target, action["value"]))



from copy import deepcopy

from rest_framework.response import Response

from ...acl.objectacl import add_acl_to_obj
from ..permissions import allow_vote_poll
from ..serializers import NewVoteSerializer, PollSerializer


def poll_vote_create(request, thread, poll):
    poll.make_choices_votes_aware(request.user)

    allow_vote_poll(request.user_acl, poll)

    serializer = NewVoteSerializer(
        data={"choices": request.data},
        context={"allowed_choices": poll.allowed_choices, "choices": poll.choices},
    )

    if not serializer.is_valid():
        return Response({"detail": serializer.errors["choices"][0]}, status=400)

    remove_user_votes(request.user, poll, serializer.data["choices"])
    set_new_votes(request, poll, serializer.data["choices"])

    add_acl_to_obj(request.user_acl, poll)
    serialized_poll = PollSerializer(poll).data

    poll.choices = list(map(presave_clean_choice, deepcopy(poll.choices)))
    poll.save()

    return Response(serialized_poll)


def presave_clean_choice(choice):
    del choice["selected"]
    return choice


def remove_user_votes(user, poll, final_votes):
    removed_votes = []
    for choice in poll.choices:
        if choice["selected"] and choice["hash"] not in final_votes:
            poll.votes -= 1
            choice["votes"] -= 1

            choice["selected"] = False
            removed_votes.append(choice["hash"])

    if removed_votes:
        poll.pollvote_set.filter(voter=user, choice_hash__in=removed_votes).delete()


def set_new_votes(request, poll, final_votes):
    for choice in poll.choices:
        if not choice["selected"] and choice["hash"] in final_votes:
            poll.votes += 1
            choice["votes"] += 1

            choice["selected"] = True
            poll.pollvote_set.create(
                category=poll.category,
                thread=poll.thread,
                voter=request.user,
                voter_name=request.user.username,
                voter_slug=request.user.slug,
                choice_hash=choice["hash"],
            )



from datetime import timedelta

import pytest
from django.utils import timezone

from ..models import WatchedThread
from ..tasks import delete_duplicate_watched_threads


def test_delete_duplicate_watched_threads_deletes_duplicates_ordered_by_read_at_desc(
    user, thread
):
    kept_watched_thread = WatchedThread.objects.create(
        user=user,
        category=thread.category,
        thread=thread,
        read_at=timezone.now(),
    )
    deleted_watched_thread = WatchedThread.objects.create(
        user=user,
        category=thread.category,
        thread=thread,
        read_at=timezone.now() - timedelta(seconds=30),
    )

    delete_duplicate_watched_threads(thread.id)

    kept_watched_thread.refresh_from_db()

    with pytest.raises(WatchedThread.DoesNotExist):
        deleted_watched_thread.refresh_from_db()

    assert WatchedThread.objects.count() == 1


def test_delete_duplicate_watched_threads_keeps_email_notifications(user, thread):
    kept_watched_thread = WatchedThread.objects.create(
        user=user,
        category=thread.category,
        thread=thread,
        send_emails=False,
        read_at=timezone.now(),
    )
    deleted_watched_thread = WatchedThread.objects.create(
        user=user,
        category=thread.category,
        thread=thread,
        send_emails=True,
        read_at=timezone.now() - timedelta(seconds=30),
    )

    delete_duplicate_watched_threads(thread.id)

    kept_watched_thread.refresh_from_db()
    assert kept_watched_thread.send_emails

    with pytest.raises(WatchedThread.DoesNotExist):
        deleted_watched_thread.refresh_from_db()

    assert WatchedThread.objects.count() == 1


def test_delete_duplicate_watched_threads_skips_non_duplicated_watched_threads(
    user, other_user, thread
):
    kept_watched_thread = WatchedThread.objects.create(
        user=user,
        category=thread.category,
        thread=thread,
        read_at=timezone.now(),
    )
    deleted_watched_thread = WatchedThread.objects.create(
        user=user,
        category=thread.category,
        thread=thread,
        read_at=timezone.now() - timedelta(seconds=30),
    )

    other_kept_watched_thread = WatchedThread.objects.create(
        user=other_user,
        category=thread.category,
        thread=thread,
        read_at=timezone.now(),
    )

    delete_duplicate_watched_threads(thread.id)

    kept_watched_thread.refresh_from_db()
    other_kept_watched_thread.refresh_from_db()

    with pytest.raises(WatchedThread.DoesNotExist):
        deleted_watched_thread.refresh_from_db()

    assert WatchedThread.objects.count() == 2


def test_delete_duplicate_watched_threads_skips_other_threads_watched_threads(
    user, other_user, thread, other_thread
):
    kept_watched_thread = WatchedThread.objects.create(
        user=user,
        category=thread.category,
        thread=thread,
        read_at=timezone.now(),
    )
    deleted_watched_thread = WatchedThread.objects.create(
        user=user,
        category=thread.category,
        thread=thread,
        read_at=timezone.now() - timedelta(seconds=30),
    )

    other_kept_watched_thread = WatchedThread.objects.create(
        user=other_user,
        category=thread.category,
        thread=thread,
        read_at=timezone.now(),
    )

    other_thread_watched_thread = WatchedThread.objects.create(
        user=other_user,
        category=other_thread.category,
        thread=other_thread,
        read_at=timezone.now(),
    )
    other_thread_watched_thread_duplicate = WatchedThread.objects.create(
        user=other_user,
        category=other_thread.category,
        thread=other_thread,
        read_at=timezone.now(),
    )

    delete_duplicate_watched_threads(thread.id)

    other_thread_watched_thread.refresh_from_db()
    other_thread_watched_thread_duplicate.refresh_from_db()

    kept_watched_thread.refresh_from_db()
    other_kept_watched_thread.refresh_from_db()

    with pytest.raises(WatchedThread.DoesNotExist):
        deleted_watched_thread.refresh_from_db()

    assert WatchedThread.objects.count() == 4



import pytest

from ..cursorpagination import CursorPage, EmptyPage, InvalidPage, get_page


@pytest.fixture
def mock_objects(mocker):
    return [mocker.Mock(post=i) for i in range(1, 12)]


@pytest.fixture
def mock_queryset(mocker, mock_objects):
    return mocker.Mock(filter=mocker.Mock(return_value=mock_objects))


def test_paginator_returns_first_page(mock_objects):
    assert get_page(mock_objects, "post", 6)


def test_first_page_has_first_flag(mock_objects):
    page = get_page(mock_objects, "post", 6)
    assert page.first


def test_first_page_start_is_zero(mock_objects):
    page = get_page(mock_objects, "post", 6)
    assert page.start == 0


def test_first_page_has_correct_length(mock_objects):
    page = get_page(mock_objects, "post", 6)
    assert len(page.object_list) == 6


def test_first_page_has_correct_items(mock_objects):
    page = get_page(mock_objects, "post", 6)
    assert page.object_list == mock_objects[:6]


def test_page_has_next_attr_pointing_to_first_item_of_next_page(mock_objects):
    page = get_page(mock_objects, "post", 6)
    assert page.next == 7


def test_requesting_next_page_filters_queryset_using_filter_name(mock_queryset):
    page = get_page(mock_queryset, "post", 6, 7)
    mock_queryset.filter.assert_called_once_with(post__gte=7)


def test_requesting_next_page_for_reversed_order_filters_queryset_with_descending(
    mock_queryset,
):
    page = get_page(mock_queryset, "-post", 6, 7)
    mock_queryset.filter.assert_called_once_with(post__lte=7)


def test_requesting_next_page_limits_queryset_to_specified_length(mock_queryset):
    page = get_page(mock_queryset, "post", 6, 7)
    assert len(page.object_list) == 6


def test_paginator_returns_empty_first_page_without_errors():
    get_page([], "post", 6)


def test_paginator_raises_empty_page_error_if_nth_page_is_empty(mocker):
    queryset = mocker.Mock(filter=lambda **_: [])
    with pytest.raises(EmptyPage):
        get_page(queryset, "post", 6, 20)


def test_paginator_raises_invalid_page_error_if_starting_position_is_negative():
    with pytest.raises(InvalidPage):
        get_page(None, None, 0, -1)


def test_page_can_be_tested_to_see_if_next_page_exists(mock_objects):
    page = get_page(mock_objects, "post", 6)
    assert page.has_next()


def test_last_page_has_no_next(mock_objects):
    page = get_page([], "post", 6)
    assert not page.next
    assert not page.has_next()


def test_cursor_page_is_first_if_start_is_zero():
    page = CursorPage(0, [])
    assert page.first


def test_cursor_page_is_not_first_if_start_is_not_zero():
    page = CursorPage(1, [])
    assert not page.first



import pytest

from ..menuitems import get_footer_menu_items, get_navbar_menu_items
from ..models import MenuItem


@pytest.fixture
def navar_menu_item(db):
    return MenuItem.objects.create(
        title="Top Menu Item",
        url="https://navbar_menu_item.com",
        menu=MenuItem.MENU_NAVBAR,
    )


@pytest.fixture
def footer_menu_item(db):
    return MenuItem.objects.create(
        title="Footer Menu Item",
        url="https://footer_menu_item.com",
        menu=MenuItem.MENU_FOOTER,
    )


@pytest.fixture
def both_menus_item(db):
    return MenuItem.objects.create(
        title="Both Positions Menu Item",
        url="https://both_menus_menu_item.com",
        menu=MenuItem.MENU_BOTH,
    )


@pytest.fixture
def menu_item_with_attributes(db):
    return MenuItem.objects.create(
        title="Menu item with attributes",
        url="https://menu_item_with_attributes.com",
        menu=MenuItem.MENU_BOTH,
        rel="noopener nofollow",
        target_blank=True,
        css_class="test-item-css-class",
    )


@pytest.fixture
def navbar_menu_items(
    db, cache_versions, navar_menu_item, both_menus_item, menu_item_with_attributes
):
    return get_navbar_menu_items(cache_versions)


@pytest.fixture
def footer_menu_items(
    db, cache_versions, footer_menu_item, both_menus_item, menu_item_with_attributes
):
    return get_footer_menu_items(cache_versions)



import random
import time

from django.contrib.auth import get_user_model
from django.core.management.base import BaseCommand

from ....core.management.progressbar import show_progress
from ....notifications.models import WatchedThread
from ....threads.models import Thread

User = get_user_model()


class Command(BaseCommand):
    help = "Adds watched threads entries for all users and threads"

    def handle(self, *args, **options):
        total_users = User.objects.count()
        total_threads = Thread.objects.count()

        self.stdout.write(
            f"Creating watched thread entries for {total_users} users "
            f"and {total_threads} threads...\n"
        )

        WatchedThread.objects.all().delete()

        total_watched_threads = total_users * total_threads
        processed_count = 0

        start_time = time.time()
        show_progress(self, processed_count, total_watched_threads)

        for user in User.objects.iterator():
            for thread in Thread.objects.iterator():
                WatchedThread.objects.create(
                    user=user,
                    category_id=thread.category_id,
                    thread=thread,
                    send_emails=random.choice((True, False)),
                    read_at=thread.last_post_on,
                )

                processed_count += 1
                show_progress(self, processed_count, total_watched_threads)

        total_time = time.time() - start_time
        total_humanized = time.strftime("%H:%M:%S", time.gmtime(total_time))
        message = "\nSuccessfully created %s watched threads in %s"
        self.stdout.write(message % (total_watched_threads, total_humanized))



from django.shortcuts import redirect


def forum_index(request):
    return  # blow up as this view is normally non-reachable!


def home_redirect(*args, **kwargs):
    return redirect("misago:index")



from django import forms
from django.utils.translation import pgettext, pgettext_lazy

from ..models import AttachmentType


def get_searchable_filetypes():
    choices = [(0, pgettext_lazy("admin attachments type filter choice", "All types"))]
    choices += [(a.id, a.name) for a in AttachmentType.objects.order_by("name")]
    return choices


class FilterAttachmentsForm(forms.Form):
    uploader = forms.CharField(
        label=pgettext_lazy("admin attachments filter form", "Uploader name contains"),
        required=False,
    )
    filename = forms.CharField(
        label=pgettext_lazy("admin attachments filter form", "Filename contains"),
        required=False,
    )
    filetype = forms.TypedChoiceField(
        label=pgettext_lazy("admin attachments filter form", "File type"),
        coerce=int,
        choices=get_searchable_filetypes,
        empty_value=0,
        required=False,
    )
    is_orphan = forms.ChoiceField(
        label=pgettext_lazy("admin attachments filter form", "State"),
        required=False,
        choices=[
            (
                "",
                pgettext_lazy(
                    "admin attachments orphan filter choice",
                    "All",
                ),
            ),
            (
                "yes",
                pgettext_lazy(
                    "admin attachments orphan filter choice",
                    "Only orphaned",
                ),
            ),
            (
                "no",
                pgettext_lazy(
                    "admin attachments orphan filter choice",
                    "Not orphaned",
                ),
            ),
        ],
    )

    def filter_queryset(self, criteria, queryset):
        if criteria.get("uploader"):
            queryset = queryset.filter(
                uploader_slug__contains=criteria["uploader"].lower()
            )
        if criteria.get("filename"):
            queryset = queryset.filter(filename__icontains=criteria["filename"])
        if criteria.get("filetype"):
            queryset = queryset.filter(filetype_id=criteria["filetype"])
        if criteria.get("is_orphan") == "yes":
            queryset = queryset.filter(post__isnull=True)
        elif criteria.get("is_orphan") == "no":
            queryset = queryset.filter(post__isnull=False)
        return queryset


class AttachmentTypeForm(forms.ModelForm):
    class Meta:
        model = AttachmentType
        fields = [
            "name",
            "extensions",
            "mimetypes",
            "size_limit",
            "status",
            "limit_uploads_to",
            "limit_downloads_to",
        ]
        labels = {
            "name": pgettext_lazy("admin attachment type form", "Type name"),
            "extensions": pgettext_lazy(
                "admin attachment type form", "File extensions"
            ),
            "mimetypes": pgettext_lazy("admin attachment type form", "Mimetypes"),
            "size_limit": pgettext_lazy(
                "admin attachment type form", "Maximum allowed uploaded file size"
            ),
            "status": pgettext_lazy("admin attachment type form", "Status"),
            "limit_uploads_to": pgettext_lazy(
                "admin attachment type form", "Limit uploads to"
            ),
            "limit_downloads_to": pgettext_lazy(
                "admin attachment type form", "Limit downloads to"
            ),
        }
        help_texts = {
            "extensions": pgettext_lazy(
                "admin attachment type form",
                "List of comma separated file extensions associated with this attachment type.",
            ),
            "mimetypes": pgettext_lazy(
                "admin attachment type form",
                "Optional list of comma separated mime types associated with this attachment type.",
            ),
            "size_limit": pgettext_lazy(
                "admin attachment type form",
                "Maximum allowed uploaded file size for this type, in kb. This setting is deprecated and has no effect. It will be deleted in Misago 1.0.",
            ),
            "status": pgettext_lazy(
                "admin attachment type form",
                "Controls this attachment type availability on your site.",
            ),
            "limit_uploads_to": pgettext_lazy(
                "admin attachment type form",
                "If you wish to limit option to upload files of this type to users with specific roles, select them on this list. Otherwise don't select any roles to allow all users with permission to upload attachments to be able to upload attachments of this type.",
            ),
            "limit_downloads_to": pgettext_lazy(
                "admin attachment type form",
                "If you wish to limit option to download files of this type to users with specific roles, select them on this list. Otherwise don't select any roles to allow all users with permission to download attachments to be able to download attachments of this type.",
            ),
        }
        widgets = {
            "limit_uploads_to": forms.CheckboxSelectMultiple,
            "limit_downloads_to": forms.CheckboxSelectMultiple,
        }

    def clean_extensions(self):
        data = self.clean_list(self.cleaned_data["extensions"])
        if not data:
            raise forms.ValidationError(
                pgettext("admin attachment type form", "This field is required.")
            )
        return data

    def clean_mimetypes(self):
        data = self.cleaned_data["mimetypes"]
        if data:
            return self.clean_list(data)

    def clean_list(self, value):
        items = [v.lstrip(".") for v in value.lower().replace(" ", "").split(",")]
        return ",".join(set(filter(bool, items)))



from ..deletesrecord import (
    record_user_deleted_by_self,
    record_user_deleted_by_staff,
    record_user_deleted_by_system,
)
from ..models import DeletedUser


def test_deletion_by_self_creates_record(db):
    record_user_deleted_by_self()
    DeletedUser.objects.get(deleted_by=DeletedUser.DELETED_BY_SELF)


def test_deletion_by_staff_creates_record(db):
    record_user_deleted_by_staff()
    DeletedUser.objects.get(deleted_by=DeletedUser.DELETED_BY_STAFF)


def test_deletion_by_system_creates_record(db):
    record_user_deleted_by_system()
    DeletedUser.objects.get(deleted_by=DeletedUser.DELETED_BY_SYSTEM)



from django.contrib.auth import get_user_model
from django.test import TestCase

from .models import AnonymousUser, Online

User = get_user_model()


class UserTestCase(TestCase):
    USER_PASSWORD = "Pass.123"
    USER_IP = "127.0.0.1"

    def setUp(self):
        super().setUp()
        self.get_initial_user()

    def get_initial_user(self):
        self.user = self.get_anonymous_user()

    def get_anonymous_user(self):
        return AnonymousUser()

    def get_authenticated_user(self):
        return create_test_user(
            "TestUser", "test@user.com", self.USER_PASSWORD, joined_from_ip=self.USER_IP
        )

    def get_superuser(self):
        return create_test_superuser(
            "TestSuperUser",
            "test@superuser.com",
            self.USER_PASSWORD,
            joined_from_ip=self.USER_IP,
        )

    def login_user(self, user, password=None):
        self.client.force_login(user)

    def logout_user(self):
        if self.user.is_authenticated:
            Online.objects.filter(user=self.user).delete()
        self.client.logout()


class AuthenticatedUserTestCase(UserTestCase):
    def get_initial_user(self):
        self.user = self.get_authenticated_user()
        self.login_user(self.user)

    def reload_user(self):
        self.user.refresh_from_db()


class SuperUserTestCase(AuthenticatedUserTestCase):
    def get_initial_user(self):
        self.user = self.get_superuser()
        self.login_user(self.user)


def create_test_user(username, email, password=None, **extra_fields):
    """Faster alternative to regular `create_user` followed by `setup_new_user`"""
    if "avatars" not in extra_fields:
        extra_fields["avatars"] = user_placeholder_avatars

    return User.objects.create_user(username, email, password, **extra_fields)


def create_test_superuser(username, email, password=None, **extra_fields):
    """Faster alternative to regular `create_superuser` followed by `setup_new_user`"""
    if "avatars" not in extra_fields:
        extra_fields["avatars"] = user_placeholder_avatars

    return User.objects.create_superuser(username, email, password, **extra_fields)


user_placeholder_avatars = [
    {"size": 400, "url": "http://dummyimage.com/400/400"},
    {"size": 200, "url": "http://dummyimage.com/200/200"},
    {"size": 100, "url": "http://dummyimage.com/100/100"},
]



from django.apps import AppConfig


class MisagoCategoriesConfig(AppConfig):
    name = "misago.categories"
    label = "misago_categories"
    verbose_name = "Misago Categories"

    def ready(self):
        from . import signals as _



from datetime import timedelta

from django.contrib.auth import get_user_model
from django.utils import timezone

from ..acl.test import patch_user_acl
from ..categories.models import Category
from ..core.utils import slugify
from ..users.test import create_test_user
from .checksums import update_post_checksum
from .models import Poll, Post, Thread

default_category_acl = {
    "can_see": 1,
    "can_browse": 1,
    "can_see_all_threads": 1,
    "can_see_own_threads": 0,
    "can_hide_threads": 0,
    "can_approve_content": 0,
    "can_edit_posts": 0,
    "can_hide_posts": 0,
    "can_hide_own_posts": 0,
    "can_merge_threads": 0,
    "can_close_threads": 0,
}


def patch_category_acl(acl_patch=None):
    def patch_acl(_, user_acl):
        category = Category.objects.get(slug="first-category")
        category_acl = user_acl["categories"][category.id]
        category_acl.update(default_category_acl)
        if acl_patch:
            category_acl.update(acl_patch)
        cleanup_patched_acl(user_acl, category_acl, category)

    return patch_user_acl(patch_acl)


def patch_other_user_category_acl(acl_patch=None):
    def patch_acl(user, user_acl):
        if user.slug != "other_user":
            return

        category = Category.objects.get(slug="first-category")
        category_acl = user_acl["categories"][category.id]
        category_acl.update(default_category_acl)
        if acl_patch:
            category_acl.update(acl_patch)
        cleanup_patched_acl(user_acl, category_acl, category)

    return patch_user_acl(patch_acl)


def patch_other_category_acl(acl_patch=None):
    def patch_acl(_, user_acl):
        src_category = Category.objects.get(slug="first-category")
        category_acl = user_acl["categories"][src_category.id].copy()

        dst_category = Category.objects.get(slug="other-category")
        user_acl["categories"][dst_category.id] = category_acl

        category_acl.update(default_category_acl)
        if acl_patch:
            category_acl.update(acl_patch)

        cleanup_patched_acl(user_acl, category_acl, dst_category)

    return patch_user_acl(patch_acl)


def patch_private_threads_acl(acl_patch=None):
    def patch_acl(_, user_acl):
        category = Category.objects.private_threads()
        category_acl = user_acl["categories"][category.id]
        category_acl.update(default_category_acl)
        if acl_patch:
            category_acl.update(acl_patch)
        cleanup_patched_acl(user_acl, category_acl, category)

    return patch_user_acl(patch_acl)


def other_user_cant_use_private_threads(user, user_acl):
    if user.slug == "other-user":
        user_acl.update({"can_use_private_threads": False})


def create_category_acl_patch(category_slug, acl_patch):
    def created_category_acl_patch(_, user_acl):
        category = Category.objects.get(slug=category_slug)
        category_acl = user_acl["categories"].get(category.id, {})
        category_acl.update(default_category_acl)
        if acl_patch:
            category_acl.update(acl_patch)
        cleanup_patched_acl(user_acl, category_acl, category)

    return created_category_acl_patch


def cleanup_patched_acl(user_acl, category_acl, category):
    visible_categories = user_acl["visible_categories"]
    browseable_categories = user_acl["browseable_categories"]

    if not category_acl["can_see"] and category.id in visible_categories:
        visible_categories.remove(category.id)

    if not category_acl["can_see"] and category.id in browseable_categories:
        browseable_categories.remove(category.id)

    if not category_acl["can_browse"] and category.id in browseable_categories:
        browseable_categories.remove(category.id)

    if category_acl["can_see"] and category.id not in visible_categories:
        visible_categories.append(category.id)

    if category_acl["can_browse"] and category.id not in browseable_categories:
        browseable_categories.append(category.id)


User = get_user_model()


def post_thread(
    category,
    title="Test thread",
    poster="Tester",
    is_global=False,
    is_pinned=False,
    is_unapproved=False,
    is_hidden=False,
    is_closed=False,
    started_on=None,
):
    started_on = started_on or timezone.now()

    kwargs = {
        "category": category,
        "title": title,
        "slug": slugify(title),
        "started_on": started_on,
        "last_post_on": started_on,
        "is_unapproved": is_unapproved,
        "is_hidden": is_hidden,
        "is_closed": is_closed,
    }

    if is_global:
        kwargs["weight"] = 2
    elif is_pinned:
        kwargs["weight"] = 1

    try:
        kwargs.update(
            {
                "starter": poster,
                "starter_name": poster.username,
                "starter_slug": poster.slug,
                "last_poster": poster,
                "last_poster_name": poster.username,
                "last_poster_slug": poster.slug,
            }
        )
    except AttributeError:
        kwargs.update(
            {
                "starter_name": poster,
                "starter_slug": slugify(poster),
                "last_poster_name": poster,
                "last_poster_slug": slugify(poster),
            }
        )

    thread = Thread.objects.create(**kwargs)
    reply_thread(
        thread,
        poster=poster,
        posted_on=started_on,
        is_hidden=is_hidden,
        is_unapproved=is_unapproved,
    )

    return thread


def reply_thread(
    thread,
    poster="Tester",
    message="I am test message",
    is_unapproved=False,
    is_hidden=False,
    is_event=False,
    is_protected=False,
    has_reports=False,
    has_open_reports=False,
    posted_on=None,
):
    posted_on = posted_on or thread.last_post_on + timedelta(minutes=5)

    kwargs = {
        "category": thread.category,
        "thread": thread,
        "original": message,
        "parsed": message,
        "checksum": "nope",
        "posted_on": posted_on,
        "updated_on": posted_on,
        "is_event": is_event,
        "is_unapproved": is_unapproved,
        "is_hidden": is_hidden,
        "is_protected": is_protected,
        "has_reports": has_reports,
        "has_open_reports": has_open_reports,
    }

    try:
        kwargs.update({"poster": poster, "poster_name": poster.username})
    except AttributeError:
        kwargs.update({"poster_name": poster})

    post = Post.objects.create(**kwargs)

    update_post_checksum(post)
    post.save()

    thread.synchronize()
    thread.save()
    thread.category.synchronize()
    thread.category.save()

    return post


def post_poll(thread, poster):
    poll = Poll.objects.create(
        category=thread.category,
        thread=thread,
        poster=poster,
        poster_name=poster.username,
        poster_slug=poster.slug,
        question="Lorem ipsum dolor met?",
        choices=[
            {"hash": "aaaaaaaaaaaa", "label": "Alpha", "votes": 1},
            {"hash": "bbbbbbbbbbbb", "label": "Beta", "votes": 0},
            {"hash": "gggggggggggg", "label": "Gamma", "votes": 2},
            {"hash": "dddddddddddd", "label": "Delta", "votes": 1},
        ],
        allowed_choices=2,
        votes=4,
    )

    # one user voted for Alpha choice
    try:
        user = User.objects.get(slug="user")
    except User.DoesNotExist:
        user = create_test_user("User", "user@example.com")

    poll.pollvote_set.create(
        category=thread.category,
        thread=thread,
        voter=user,
        voter_name=user.username,
        voter_slug=user.slug,
        choice_hash="aaaaaaaaaaaa",
    )

    # test user voted on third and last choices
    poll.pollvote_set.create(
        category=thread.category,
        thread=thread,
        voter=poster,
        voter_name=poster.username,
        voter_slug=poster.slug,
        choice_hash="gggggggggggg",
    )
    poll.pollvote_set.create(
        category=thread.category,
        thread=thread,
        voter=poster,
        voter_name=poster.username,
        voter_slug=poster.slug,
        choice_hash="dddddddddddd",
    )

    # somebody else voted on third option before being deleted
    poll.pollvote_set.create(
        category=thread.category,
        thread=thread,
        voter_name="deleted",
        voter_slug="deleted",
        choice_hash="gggggggggggg",
    )

    return poll


def like_post(post, liker=None, username=None):
    if not post.last_likes:
        post.last_likes = []

    if liker:
        like = post.postlike_set.create(
            category=post.category,
            thread=post.thread,
            liker=liker,
            liker_name=liker.username,
            liker_slug=liker.slug,
        )

        post.last_likes = [
            {"id": liker.id, "username": liker.username}
        ] + post.last_likes
    else:
        like = post.postlike_set.create(
            category=post.category,
            thread=post.thread,
            liker_name=username,
            liker_slug=slugify(username),
        )

        post.last_likes = [{"id": None, "username": username}] + post.last_likes

    post.likes += 1
    post.save()

    return like



from enum import Enum


class PluginOutlet(Enum):
    """Enum with standard plugin outlets defined by Misago.

    Members values are descriptions of outlets locations, used by the docs generator.
    """

    TEST = "Used in some tests."
    ADMIN_DASHBOARD_START = "On the admin dashboard page, above all other content."
    ADMIN_DASHBOARD_AFTER_CHECKS = "On the admin dashboard page, below the Checks card."
    ADMIN_DASHBOARD_AFTER_ANALYTICS = (
        "On the admin dashboard page, below the Analytics card."
    )
    ADMIN_DASHBOARD_END = "On the admin dashboard page, below all other content."
    CATEGORIES_LIST_START = "On the categories page, above the list."
    CATEGORIES_LIST_END = "On the categories page, under the list."



from unittest.mock import patch

from django.urls import reverse

from .. import test
from ...acl.test import patch_user_acl
from ...categories.models import Category
from ...users.test import AuthenticatedUserTestCase
from ..models import Thread
from ..test import patch_category_acl


class ReplyThreadTests(AuthenticatedUserTestCase):
    def setUp(self):
        super().setUp()

        self.category = Category.objects.get(slug="first-category")
        self.thread = test.post_thread(category=self.category)

        self.api_link = reverse(
            "misago:api:thread-post-list", kwargs={"thread_pk": self.thread.pk}
        )

    def test_cant_reply_thread_as_guest(self):
        """user has to be authenticated to be able to post reply"""
        self.logout_user()

        response = self.client.post(self.api_link)
        self.assertEqual(response.status_code, 403)

    def test_thread_visibility(self):
        """thread's visibility is validated"""
        with patch_category_acl({"can_see": 0}):
            response = self.client.post(self.api_link)
            self.assertEqual(response.status_code, 404)

        with patch_category_acl({"can_browse": 0}):
            response = self.client.post(self.api_link)
            self.assertEqual(response.status_code, 404)

        with patch_category_acl({"can_see_all_threads": 0}):
            response = self.client.post(self.api_link)
            self.assertEqual(response.status_code, 404)

    @patch_category_acl({"can_reply_threads": False})
    def test_cant_reply_thread(self):
        """permission to reply thread is validated"""
        response = self.client.post(self.api_link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(), {"detail": "You can't reply to threads in this category."}
        )

    @patch_category_acl({"can_reply_threads": True, "can_close_threads": False})
    def test_closed_category_no_permission(self):
        """permssion to reply in closed category is validated"""
        self.category.is_closed = True
        self.category.save()

        response = self.client.post(self.api_link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(),
            {"detail": "This category is closed. You can't reply to threads in it."},
        )

    @patch_category_acl({"can_reply_threads": True, "can_close_threads": True})
    def test_closed_category(self):
        """permssion to reply in closed category is validated"""
        self.category.is_closed = True
        self.category.save()

        response = self.client.post(self.api_link)
        self.assertEqual(response.status_code, 400)

    @patch_category_acl({"can_reply_threads": True, "can_close_threads": False})
    def test_closed_thread_no_permission(self):
        """permssion to reply in closed thread is validated"""
        self.thread.is_closed = True
        self.thread.save()

        response = self.client.post(self.api_link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(),
            {"detail": "You can't reply to closed threads in this category."},
        )

    @patch_category_acl({"can_reply_threads": True, "can_close_threads": True})
    def test_closed_thread(self):
        """permssion to reply in closed thread is validated"""
        self.thread.is_closed = True
        self.thread.save()

        response = self.client.post(self.api_link)
        self.assertEqual(response.status_code, 400)

    @patch_category_acl({"can_reply_threads": True})
    def test_empty_data(self):
        """no data sent handling has no showstoppers"""
        response = self.client.post(self.api_link, data={})
        self.assertEqual(response.status_code, 400)
        self.assertEqual(response.json(), {"post": ["You have to enter a message."]})

    @patch_category_acl({"can_reply_threads": True})
    def test_invalid_data(self):
        """api errors for invalid request data"""
        response = self.client.post(
            self.api_link, "false", content_type="application/json"
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "non_field_errors": [
                    "Invalid data. Expected a dictionary, but got bool."
                ]
            },
        )

    @patch_category_acl({"can_reply_threads": True})
    def test_post_is_validated(self):
        """post is validated"""
        response = self.client.post(self.api_link, data={"post": "a"})

        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "post": [
                    "Posted message should be at least 5 characters long (it has 1)."
                ]
            },
        )

    @patch_category_acl({"can_reply_threads": True})
    @patch(
        "misago.threads.api.postingendpoint.notifications.notify_on_new_thread_reply"
    )
    def test_can_reply_thread(self, notify_on_new_thread_reply_mock):
        """endpoint creates new reply"""
        response = self.client.post(
            self.api_link, data={"post": "This is test response!"}
        )
        self.assertEqual(response.status_code, 200)

        thread = Thread.objects.get(pk=self.thread.pk)

        response = self.client.get(self.thread.get_absolute_url())
        self.assertContains(response, "<p>This is test response!</p>")

        # api increased user's posts counts
        self.reload_user()
        self.assertEqual(self.user.threads, 0)
        self.assertEqual(self.user.posts, 1)

        self.assertEqual(self.user.audittrail_set.count(), 1)

        post = self.user.post_set.all()[:1][0]
        self.assertEqual(post.category_id, self.category.pk)
        self.assertEqual(post.original, "This is test response!")
        self.assertEqual(post.poster_id, self.user.id)
        self.assertEqual(post.poster_name, self.user.username)

        self.assertEqual(thread.last_post_id, post.id)
        self.assertEqual(thread.last_poster_id, self.user.id)
        self.assertEqual(thread.last_poster_name, self.user.username)
        self.assertEqual(thread.last_poster_slug, self.user.slug)

        category = Category.objects.get(pk=self.category.pk)
        self.assertEqual(category.last_thread_id, thread.id)
        self.assertEqual(category.last_thread_title, thread.title)
        self.assertEqual(category.last_thread_slug, thread.slug)

        self.assertEqual(category.last_poster_id, self.user.id)
        self.assertEqual(category.last_poster_name, self.user.username)
        self.assertEqual(category.last_poster_slug, self.user.slug)

    @patch_category_acl({"can_reply_threads": True})
    @patch(
        "misago.threads.api.postingendpoint.notifications.notify_on_new_thread_reply"
    )
    def test_post_unicode(self, notify_on_new_thread_reply_mock):
        """unicode characters can be posted"""
        response = self.client.post(
            self.api_link, data={"post": "Chrzążczyżewoszyce, powiat Łękółody."}
        )
        self.assertEqual(response.status_code, 200)

    @patch_category_acl({"can_reply_threads": True})
    @patch(
        "misago.threads.api.postingendpoint.notifications.notify_on_new_thread_reply"
    )
    def test_category_moderation_queue(self, notify_on_new_thread_reply_mock):
        """reply thread in category that requires approval"""
        self.category.require_replies_approval = True
        self.category.save()

        response = self.client.post(
            self.api_link, data={"post": "Lorem ipsum dolor met!"}
        )
        self.assertEqual(response.status_code, 200)

        thread = Thread.objects.get(pk=self.thread.pk)
        self.assertFalse(thread.is_unapproved)
        self.assertTrue(thread.has_unapproved_posts)
        self.assertEqual(thread.replies, self.thread.replies)

        post = self.user.post_set.all()[:1][0]
        self.assertTrue(post.is_unapproved)

        category = Category.objects.get(slug="first-category")
        self.assertEqual(category.threads, self.category.threads)
        self.assertEqual(category.posts, self.category.posts)

    @patch_category_acl({"can_reply_threads": True})
    @patch_user_acl({"can_approve_content": True})
    @patch(
        "misago.threads.api.postingendpoint.notifications.notify_on_new_thread_reply"
    )
    def test_category_moderation_queue_bypass(self, notify_on_new_thread_reply_mock):
        """bypass moderation queue due to user's acl"""
        self.category.require_replies_approval = True
        self.category.save()

        response = self.client.post(
            self.api_link, data={"post": "Lorem ipsum dolor met!"}
        )
        self.assertEqual(response.status_code, 200)

        thread = Thread.objects.get(pk=self.thread.pk)
        self.assertFalse(thread.is_unapproved)
        self.assertFalse(thread.has_unapproved_posts)
        self.assertEqual(thread.replies, self.thread.replies + 1)

        post = self.user.post_set.all()[:1][0]
        self.assertFalse(post.is_unapproved)

        category = Category.objects.get(slug="first-category")
        self.assertEqual(category.threads, self.category.threads)
        self.assertEqual(category.posts, self.category.posts + 1)

    @patch_category_acl({"can_reply_threads": True, "require_replies_approval": True})
    @patch(
        "misago.threads.api.postingendpoint.notifications.notify_on_new_thread_reply"
    )
    def test_user_moderation_queue(self, notify_on_new_thread_reply_mock):
        """reply thread by user that requires approval"""
        response = self.client.post(
            self.api_link, data={"post": "Lorem ipsum dolor met!"}
        )
        self.assertEqual(response.status_code, 200)

        thread = Thread.objects.get(pk=self.thread.pk)
        self.assertFalse(thread.is_unapproved)
        self.assertTrue(thread.has_unapproved_posts)
        self.assertEqual(thread.replies, self.thread.replies)

        post = self.user.post_set.all()[:1][0]
        self.assertTrue(post.is_unapproved)

        category = Category.objects.get(slug="first-category")
        self.assertEqual(category.threads, self.category.threads)
        self.assertEqual(category.posts, self.category.posts)

    @patch_category_acl({"can_reply_threads": True, "require_replies_approval": True})
    @patch_user_acl({"can_approve_content": True})
    @patch(
        "misago.threads.api.postingendpoint.notifications.notify_on_new_thread_reply"
    )
    def test_user_moderation_queue_bypass(self, notify_on_new_thread_reply_mock):
        """bypass moderation queue due to user's acl"""
        response = self.client.post(
            self.api_link, data={"post": "Lorem ipsum dolor met!"}
        )
        self.assertEqual(response.status_code, 200)

        thread = Thread.objects.get(pk=self.thread.pk)
        self.assertFalse(thread.is_unapproved)
        self.assertFalse(thread.has_unapproved_posts)
        self.assertEqual(thread.replies, self.thread.replies + 1)

        post = self.user.post_set.all()[:1][0]
        self.assertFalse(post.is_unapproved)

        category = Category.objects.get(slug="first-category")
        self.assertEqual(category.threads, self.category.threads)
        self.assertEqual(category.posts, self.category.posts + 1)

    @patch_category_acl(
        {
            "can_reply_threads": True,
            "require_threads_approval": True,
            "require_edits_approval": True,
        }
    )
    @patch(
        "misago.threads.api.postingendpoint.notifications.notify_on_new_thread_reply"
    )
    def test_omit_other_moderation_queues(self, notify_on_new_thread_reply_mock):
        """other queues are omitted"""
        self.category.require_threads_approval = True
        self.category.require_edits_approval = True
        self.category.save()

        response = self.client.post(
            self.api_link, data={"post": "Lorem ipsum dolor met!"}
        )
        self.assertEqual(response.status_code, 200)

        thread = Thread.objects.get(pk=self.thread.pk)
        self.assertFalse(thread.is_unapproved)
        self.assertFalse(thread.has_unapproved_posts)
        self.assertEqual(thread.replies, self.thread.replies + 1)

        post = self.user.post_set.all()[:1][0]
        self.assertFalse(post.is_unapproved)

        category = Category.objects.get(slug="first-category")
        self.assertEqual(category.threads, self.category.threads)
        self.assertEqual(category.posts, self.category.posts + 1)



from typing import Any

from django.contrib.postgres.indexes import GinIndex
from django.db import models


class PluginDataModel(models.Model):
    plugin_data = models.JSONField(default=dict)

    class Meta:
        abstract = True
        indexes = [GinIndex(fields=["plugin_data"])]



from rest_framework import serializers

from ..models import DataDownload

__all__ = ["DataDownloadSerializer"]


class DataDownloadSerializer(serializers.ModelSerializer):
    class Meta:
        model = DataDownload
        fields = ["id", "status", "requested_on", "expires_on", "file"]



from django.db import models, transaction
from django.urls import reverse
from django.utils.translation import pgettext

from ...acl.cache import clear_acl_cache
from ...core.utils import slugify


class RankManager(models.Manager):
    def get_default(self):
        return self.get(is_default=True)

    def make_rank_default(self, rank):
        with transaction.atomic():
            self.filter(is_default=True).update(is_default=False)
            rank.is_default = True
            rank.save(update_fields=["is_default"])


class Rank(models.Model):
    name = models.CharField(max_length=255)
    slug = models.CharField(unique=True, max_length=255)
    description = models.TextField(null=True, blank=True)
    title = models.CharField(max_length=255, null=True, blank=True)
    roles = models.ManyToManyField("misago_acl.Role", blank=True)
    css_class = models.CharField(max_length=255, null=True, blank=True)
    is_default = models.BooleanField(default=False)
    is_tab = models.BooleanField(default=False)
    order = models.IntegerField(default=0)

    objects = RankManager()

    class Meta:
        get_latest_by = "order"

    def __str__(self):
        return pgettext("rank name", self.name)

    def save(self, *args, **kwargs):
        if not self.pk:
            self.set_order()
        else:
            clear_acl_cache()
        if not self.slug:
            self.slug = slugify(self.name)
        return super().save(*args, **kwargs)

    def delete(self, *args, **kwargs):
        clear_acl_cache()
        return super().delete(*args, **kwargs)

    def get_absolute_url(self):
        return reverse("misago:users-rank", kwargs={"slug": self.slug})

    def set_name(self, name):
        self.name = name
        self.slug = slugify(name)

    def set_order(self):
        try:
            self.order = Rank.objects.latest("order").order + 1
        except Rank.DoesNotExist:
            self.order = 0



# Generated by Django 3.2.15 on 2023-01-04 12:38

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("misago_users", "0022_deleteduser"),
        ("misago_oauth2", "0002_copy_sso_subjects"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="user",
            name="sso_id",
        ),
    ]



from ..models import WatchedThread
from ..threads import ThreadNotifications, watch_started_thread


def test_started_thread_is_watched_with_email_notifications(user, thread):
    user.watch_started_threads = ThreadNotifications.SITE_AND_EMAIL
    user.save()

    watch_started_thread(user, thread)

    watched_thread = WatchedThread.objects.get(user=user, thread=thread)
    assert watched_thread.send_emails


def test_started_thread_is_watched_without_email_notifications(user, thread):
    user.watch_started_threads = ThreadNotifications.SITE_ONLY
    user.save()

    watch_started_thread(user, thread)

    watched_thread = WatchedThread.objects.get(user=user, thread=thread)
    assert not watched_thread.send_emails


def test_started_thread_is_not_watched_if_option_is_disabled(user, thread):
    user.watch_started_threads = ThreadNotifications.NONE
    user.save()

    watch_started_thread(user, thread)

    assert not WatchedThread.objects.exists()



from django.shortcuts import redirect, render as dj_render

from . import get_protected_namespace, protected_admin_view, render
from ...core.utils import get_exception_message
from ..auth import is_admin_authorized, update_admin_authorization


# Magic error page used by admin
@protected_admin_view
def _error_page(request, code, exception=None, default_message=None):
    if not is_admin_authorized(request):
        return redirect("misago:admin:index")

    template_pattern = "misago/admin/errorpages/%s.html" % code

    response = render(
        request,
        template_pattern,
        {"message": get_exception_message(exception, default_message)},
        error_page=True,
    )
    response.status_code = code
    return response


def admin_error_page(f):
    def decorator(request, *args, **kwargs):
        if get_protected_namespace(request):
            update_admin_authorization(request)
            return _error_page(request, *args, **kwargs)
        return f(request, *args, **kwargs)

    return decorator


# Magic CSRF fail page for Admin
def _csrf_failure(request, reason=""):
    if is_admin_authorized(request):
        update_admin_authorization(request)
        response = render(
            request,
            "misago/admin/errorpages/csrf_failure_authenticated.html",
            error_page=True,
        )
    else:
        response = dj_render(request, "misago/admin/errorpages/csrf_failure.html")

    response.status_code = 403
    return response


def admin_csrf_failure(f):
    def decorator(request, *args, **kwargs):
        if get_protected_namespace(request):
            return _csrf_failure(request, *args, **kwargs)
        return f(request, *args, **kwargs)

    return decorator



# Generated by Django 2.2.3 on 2019-08-15 16:58

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [("misago_users", "0020_set_dj_partial_indexes")]

    operations = [
        migrations.AddField(
            model_name="user",
            name="sso_id",
            field=models.PositiveIntegerField(blank=True, null=True, unique=True),
        )
    ]



from django.contrib import messages
from django.shortcuts import redirect
from django.urls import reverse
from django.utils.translation import pgettext_lazy

from ...admin.views import generic
from ..models import Role
from .forms import RoleForm, get_permissions_forms


class RoleAdmin(generic.AdminBaseMixin):
    root_link = "misago:admin:permissions:index"
    model = Role
    templates_dir = "misago/admin/roles"
    message_404 = pgettext_lazy("admin roles", "Requested role does not exist.")


class RolesList(RoleAdmin, generic.ListView):
    ordering = (("name", None),)


class RoleFormMixin:
    def real_dispatch(self, request, target):
        form = RoleForm(instance=target)

        perms_forms = get_permissions_forms(target)

        if request.method == "POST":
            perms_forms = get_permissions_forms(target, request.POST)
            valid_forms = 0
            for permissions_form in perms_forms:
                if permissions_form.is_valid():
                    valid_forms += 1

            form = RoleForm(request.POST, instance=target)
            if form.is_valid() and len(perms_forms) == valid_forms:
                new_permissions = {}
                for permissions_form in perms_forms:
                    cleaned_data = permissions_form.cleaned_data
                    new_permissions[permissions_form.prefix] = cleaned_data

                form.instance.permissions = new_permissions
                form.instance.save()

                messages.success(request, self.message_submit % {"name": target.name})

                if "stay" in request.POST:
                    return redirect(request.path)
                return redirect(self.root_link)
            if form.is_valid() and len(perms_forms) != valid_forms:
                form.add_error(
                    None, pgettext_lazy("admin form", "Form contains errors.")
                )

        template_name = self.get_template_name(request, target)
        return self.render(
            request,
            {"form": form, "target": target, "perms_forms": perms_forms},
            template_name,
        )


class NewRole(RoleFormMixin, RoleAdmin, generic.ModelFormView):
    message_submit = pgettext_lazy("admin roles", 'New role "%(name)s" has been saved.')


class EditRole(RoleFormMixin, RoleAdmin, generic.ModelFormView):
    message_submit = pgettext_lazy("admin roles", 'Role "%(name)s" has been changed.')


class DeleteRole(RoleAdmin, generic.ButtonView):
    def check_permissions(self, request, target):
        if target.special_role:
            message = pgettext_lazy(
                "admin roles", 'Role "%(name)s" is special role and can\'t be deleted.'
            )
            return message % {"name": target.name}

    def button_action(self, request, target):
        target.delete()
        message = pgettext_lazy("admin roles", 'Role "%(name)s" has been deleted.')
        messages.success(request, message % {"name": target.name})


class RoleUsers(RoleAdmin, generic.TargetedView):
    def real_dispatch(self, request, target):
        redirect_url = reverse("misago:admin:users:index")
        return redirect("%s?role=%s" % (redirect_url, target.pk))



from io import StringIO

from django.core.management import call_command

from ...cache.test import assert_invalidates_cache
from ...users import BANS_CACHE
from ...users.models import Ban
from ..management.commands import createfakebans


def test_management_command_creates_fake_bans(db):
    call_command(createfakebans.Command(), stdout=StringIO())
    assert Ban.objects.exists()


def test_management_command_invalidates_bans_cache(db):
    with assert_invalidates_cache(BANS_CACHE):
        call_command(createfakebans.Command(), stdout=StringIO())



from pathlib import Path
from tempfile import TemporaryDirectory

import pytest

from ..discover import discover_plugins


@pytest.fixture
def sys_mock(mocker):
    mock = mocker.patch("misago.plugins.discover.sys")
    mock.path = []
    return mock


def test_discover_plugins_returns_empty_list_if_plugins_path_is_empty(sys_mock):
    plugins = discover_plugins("")
    assert plugins == []


def test_discover_plugins_doesnt_update_sys_path_if_plugins_path_is_empty(sys_mock):
    assert not discover_plugins("")
    assert sys_mock.path == []


def test_discover_plugins_returns_empty_list_if_plugins_path_doesnt_exist(sys_mock):
    with TemporaryDirectory() as plugins_dir:
        plugins = discover_plugins(str(Path(plugins_dir) / "doesnt_exist"))
        assert plugins == []


def test_discover_plugins_doesnt_update_sys_path_if_plugins_path_doesnt_exist(sys_mock):
    with TemporaryDirectory() as plugins_dir:
        assert not discover_plugins(str(Path(plugins_dir) / "doesnt_exist"))
        assert sys_mock.path == []


def test_discover_plugins_returns_empty_list_if_plugins_dir_is_empty(sys_mock):
    with TemporaryDirectory() as plugins_dir:
        plugins = discover_plugins(plugins_dir)
        assert plugins == []


def test_discover_plugins_doesnt_update_sys_path_if_plugins_dir_is_empty(sys_mock):
    with TemporaryDirectory() as plugins_dir:
        assert not discover_plugins(plugins_dir)
        assert sys_mock.path == []


def create_plugin(plugins_dir: str, dir_name: str, package_name: str) -> Path:
    plugin_path = Path(plugins_dir) / dir_name
    plugin_path.mkdir()

    plugin_package_path = plugin_path / package_name
    plugin_package_path.mkdir()

    plugin_init_path = plugin_package_path / "__init__.py"
    plugin_init_path.touch()

    plugin_misago_marker_path = plugin_package_path / "misago_plugin.py"
    plugin_misago_marker_path.touch()

    return plugin_misago_marker_path


def test_discover_plugins_returns_plugins_list(sys_mock):
    with TemporaryDirectory() as plugins_dir:
        create_plugin(plugins_dir, "mock-plugin", "mock_plugin")
        create_plugin(plugins_dir, "other-plugin", "other_plugin")

        plugins = discover_plugins(plugins_dir)
        assert plugins == ["mock_plugin", "other_plugin"]


def test_discover_plugins_adds_plugins_to_python_path(sys_mock):
    with TemporaryDirectory() as plugins_dir:
        create_plugin(plugins_dir, "mock-plugin", "mock_plugin")
        create_plugin(plugins_dir, "other-plugin", "other_plugin")

        assert discover_plugins(plugins_dir)
        assert sys_mock.path == [
            f"{plugins_dir}/mock-plugin",
            f"{plugins_dir}/other-plugin",
        ]


def test_discover_plugins_doesnt_create_duplicates_in_python_path(sys_mock):
    with TemporaryDirectory() as plugins_dir:
        create_plugin(plugins_dir, "mock-plugin", "mock_plugin")
        create_plugin(plugins_dir, "other-plugin", "other_plugin")

        sys_mock.path.append(f"{plugins_dir}/mock-plugin")

        assert discover_plugins(plugins_dir)
        assert sys_mock.path == [
            f"{plugins_dir}/mock-plugin",
            f"{plugins_dir}/other-plugin",
        ]


def test_discover_plugins_skips_packages_without_misago_plugin_py(sys_mock):
    with TemporaryDirectory() as plugins_dir:
        create_plugin(plugins_dir, "mock-plugin", "mock_plugin")

        other_plugin = create_plugin(plugins_dir, "other-plugin", "other_plugin")
        other_plugin.unlink()

        plugins = discover_plugins(plugins_dir)
        assert plugins == ["mock_plugin"]
        assert sys_mock.path == [f"{plugins_dir}/mock-plugin"]


def test_discover_plugins_returns_plugins_apps_sorted_by_name(sys_mock):
    with TemporaryDirectory() as plugins_dir:
        create_plugin(plugins_dir, "mock-plugin", "mock_plugin")
        create_plugin(plugins_dir, "other-plugin", "alpha_plugin")

        plugins = discover_plugins(plugins_dir)
        assert plugins == ["alpha_plugin", "mock_plugin"]


def test_discover_plugins_adds_plugins_to_python_path_ordered_by_app_name(sys_mock):
    with TemporaryDirectory() as plugins_dir:
        create_plugin(plugins_dir, "mock-plugin", "mock_plugin")
        create_plugin(plugins_dir, "other-plugin", "alpha_plugin")

        assert discover_plugins(plugins_dir)
        assert sys_mock.path == [
            f"{plugins_dir}/other-plugin",
            f"{plugins_dir}/mock-plugin",
        ]


def test_discover_plugins_skips_plugins_without_init(sys_mock):
    with TemporaryDirectory() as plugins_dir:
        create_plugin(plugins_dir, "mock-plugin", "mock_plugin")

        invalid_plugin_path = create_plugin(plugins_dir, "invalid-plugin", "invalid")
        invalid_plugin_path_init = invalid_plugin_path.parent / "__init__.py"
        invalid_plugin_path_init.unlink()

        plugins = discover_plugins(plugins_dir)
        assert plugins == ["mock_plugin"]
        assert sys_mock.path == [f"{plugins_dir}/mock-plugin"]


def create_pip_install(plugins_dir: str, contents: list[str]):
    with open(Path(plugins_dir) / "pip-install.txt", "w") as fp:
        fp.write("\n".join(contents))


@pytest.mark.xfail(reason="requires misago-pypi-plugin installed to pass")
def test_discover_plugins_reads_plugins_from_pip_install_if_it_exists(sys_mock):
    with TemporaryDirectory() as plugins_dir:
        create_pip_install(plugins_dir, ["misago-pypi-plugin"])

        plugins = discover_plugins(plugins_dir)
        assert plugins == ["misago_pypi_plugin"]
        assert sys_mock.path == []


@pytest.mark.xfail(reason="requires misago-pypi-plugin installed to pass")
def test_discover_plugins_skips_pip_install_entries_that_arent_installed(sys_mock):
    with TemporaryDirectory() as plugins_dir:
        create_pip_install(plugins_dir, ["misago-pypi-plugin", "misago-other-plugin"])

        plugins = discover_plugins(plugins_dir)
        assert plugins == ["misago_pypi_plugin"]
        assert sys_mock.path == []


@pytest.mark.xfail(reason="requires misago-pypi-plugin installed to pass")
def test_discover_plugins_skips_pip_install_entries_that_arent_misago_plugins(sys_mock):
    with TemporaryDirectory() as plugins_dir:
        create_pip_install(plugins_dir, ["misago-pypi-plugin", "django"])

        plugins = discover_plugins(plugins_dir)
        assert plugins == ["misago_pypi_plugin"]
        assert sys_mock.path == []


@pytest.mark.xfail(reason="requires misago-pypi-plugin installed to pass")
def test_discover_plugins_skips_pip_install_if_its_directory(sys_mock):
    with TemporaryDirectory() as plugins_dir:
        pip_install_path = Path(plugins_dir) / "pip-install.txt"
        pip_install_path.mkdir()

        plugins = discover_plugins(plugins_dir)
        assert plugins == []
        assert sys_mock.path == []



from ..models import WatchedThread
from ..threads import ThreadNotifications, watch_new_private_thread


def test_private_thread_is_watched_with_email_notifications(user, private_thread):
    user.watch_new_private_threads_by_other_users = ThreadNotifications.SITE_AND_EMAIL
    user.save()

    watched_thread = watch_new_private_thread(user, private_thread, from_followed=False)

    assert watched_thread.user == user
    assert watched_thread.category == private_thread.category
    assert watched_thread.thread == private_thread
    assert watched_thread.send_emails
    assert watched_thread.read_at < private_thread.started_on

    WatchedThread.objects.get(
        user=user,
        thread=private_thread,
        send_emails=True,
    )


def test_private_thread_is_watched_without_email_notifications(user, private_thread):
    user.watch_new_private_threads_by_other_users = ThreadNotifications.SITE_ONLY
    user.save()

    watched_thread = watch_new_private_thread(user, private_thread, from_followed=False)

    assert watched_thread.user == user
    assert watched_thread.category == private_thread.category
    assert watched_thread.thread == private_thread
    assert not watched_thread.send_emails
    assert watched_thread.read_at < private_thread.started_on

    WatchedThread.objects.get(
        user=user,
        thread=private_thread,
        send_emails=False,
    )


def test_private_thread_is_not_watched(user, private_thread):
    user.watch_new_private_threads_by_other_users = ThreadNotifications.NONE
    user.save()

    watched_thread = watch_new_private_thread(user, private_thread, from_followed=False)
    assert watched_thread is None

    assert not WatchedThread.objects.exists()


def test_private_thread_from_followed_is_watched_with_email_notifications(
    user, private_thread
):
    user.watch_new_private_threads_by_followed = ThreadNotifications.SITE_AND_EMAIL
    user.save()

    watched_thread = watch_new_private_thread(user, private_thread, from_followed=True)

    assert watched_thread.user == user
    assert watched_thread.category == private_thread.category
    assert watched_thread.thread == private_thread
    assert watched_thread.send_emails
    assert watched_thread.read_at < private_thread.started_on

    WatchedThread.objects.get(
        user=user,
        thread=private_thread,
        send_emails=True,
    )


def test_private_thread_from_followed_is_watched_without_email_notifications(
    user, private_thread
):
    user.watch_new_private_threads_by_followed = ThreadNotifications.SITE_ONLY
    user.save()

    watched_thread = watch_new_private_thread(user, private_thread, from_followed=True)

    assert watched_thread.user == user
    assert watched_thread.category == private_thread.category
    assert watched_thread.thread == private_thread
    assert not watched_thread.send_emails
    assert watched_thread.read_at < private_thread.started_on

    WatchedThread.objects.get(
        user=user,
        thread=private_thread,
        send_emails=False,
    )


def test_private_thread_from_followed_is_not_watched(user, private_thread):
    user.watch_new_private_threads_by_followed = ThreadNotifications.NONE
    user.save()

    watched_thread = watch_new_private_thread(user, private_thread, from_followed=True)
    assert watched_thread is None

    assert not WatchedThread.objects.exists()


def test_old_watched_thread_notifications_are_not_disabled_by_new_preference(
    user, private_thread
):
    old_watched_thread = WatchedThread.objects.create(
        user=user,
        thread=private_thread,
        category=private_thread.category,
        send_emails=True,
    )

    user.watch_new_private_threads_by_other_users = ThreadNotifications.SITE_ONLY
    user.save()

    watched_thread = watch_new_private_thread(user, private_thread, from_followed=False)

    assert watched_thread.id == old_watched_thread.id
    assert watched_thread.send_emails

    WatchedThread.objects.get(
        user=user,
        thread=private_thread,
        send_emails=True,
    )



# Bunch of tests for unsafe/malicious contents escaping
from ..parser import parse


def test_text_is_escaped(request_mock, user, snapshot):
    text = '<script>alert("!")</script>'
    result = parse(text, request_mock, user)
    assert snapshot == result["parsed_text"]


def test_inline_code_is_escaped(request_mock, user, snapshot):
    text = '`<script>alert("!")</script>`'
    result = parse(text, request_mock, user)
    assert snapshot == result["parsed_text"]


def test_code_in_quote_markdown_is_escaped(request_mock, user, snapshot):
    text = '> <script>alert("!")</script>'
    result = parse(text, request_mock, user)
    assert snapshot == result["parsed_text"]


def test_code_in_quote_bbcode_is_escaped(request_mock, user, snapshot):
    text = '[quote]<script>alert("!")</script>[/quote]'
    result = parse(text, request_mock, user)
    assert snapshot == result["parsed_text"]


def test_code_in_quote_bbcode_header_is_escaped(request_mock, user, snapshot):
    text = '[quote="@Us"><script>alert("!")</script>er"]Test[/quote]'
    result = parse(text, request_mock, user)
    assert snapshot == result["parsed_text"]



from ...conf.test import override_dynamic_settings
from ..datadownloads import request_user_data_download
from ..test import AuthenticatedUserTestCase


class UserRequestDataDownload(AuthenticatedUserTestCase):
    def setUp(self):
        super().setUp()
        self.link = "/api/users/%s/request-data-download/" % self.user.pk

    def test_request_other_user_download_anonymous(self):
        """requests to api fails if user is anonymous"""
        self.logout_user()

        response = self.client.post(self.link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(), {"detail": "This action is not available to guests."}
        )

    def test_request_other_user_download(self):
        """requests to api fails if user tries to access other user"""
        other_user = self.get_superuser()
        link = "/api/users/%s/request-data-download/" % other_user.pk

        response = self.client.post(link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(),
            {"detail": "You can't request data downloads for other users."},
        )

    @override_dynamic_settings(allow_data_downloads=False)
    def test_request_download_disabled(self):
        """request to api fails if own data downloads are disabled"""
        response = self.client.post(self.link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(response.json(), {"detail": "You can't download your data."})

    def test_request_download_in_progress(self):
        """request to api fails if user has already requested data download"""
        request_user_data_download(self.user)

        response = self.client.post(self.link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(),
            {
                "detail": (
                    "You can't have more than one data download request at a single time."
                )
            },
        )

    def test_request_download(self):
        """request to api succeeds"""
        response = self.client.post(self.link)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), {"detail": "ok"})

        self.assertTrue(self.user.datadownload_set.exists())



from django.test import override_settings

from ..momentjs import clean_language_name, get_locale_url


def test_django_language_code_is_cleaned_for_moment_js():
    assert clean_language_name("en", ["en"]) == "en"


def test_two_part_django_language_code_is_cleaned_for_moment_js():
    assert clean_language_name("en_us", ["en-us"]) == "en-us"


def test_two_part_django_language_code_is_fallbacked_to_single_part_code():
    assert clean_language_name("en_us", ["en", "en-uk"]) == "en"


def test_django_language_code_cleanup_is_case_insensitive():
    assert clean_language_name("en_US", ["en-us"]) == "en-us"


def test_django_language_code_is_cleaned_to_none_if_translation_is_not_available():
    assert clean_language_name("pl", ["en"]) is None


def test_locale_url_getter_uses_passed_locales_list():
    with override_settings(MISAGO_MOMENT_JS_LOCALES=["en-us"]):
        assert get_locale_url("en_us", locales=["en"]).endswith("en.js")


def test_locale_url_getter_fallbacks_to_locales_setting_if_locales_kwarg_is_not_set():
    with override_settings(MISAGO_MOMENT_JS_LOCALES=["en-us"]):
        assert get_locale_url("en_us").endswith("en-us.js")


def test_locale_url_getter_uses_passed_static_path_template():
    locale_url = get_locale_url(
        "en_us", static_path_template="test/%s", locales=["en-us"]
    )
    assert locale_url == "test/en-us"


def test_locale_url_getter_returns_none_for_unsupported_locale():
    assert get_locale_url("pl", locales=["en"]) is None



import pytest

from ..hooks import ActionHook


class MockActionHook(ActionHook):
    def __call__(self, base: str):
        return super().__call__(base)


def lowercase_action(base: str) -> str:
    return base.lower()


def uppercase_action(base: str) -> str:
    return base.upper()


def reverse_action(base: str) -> str:
    return "".join(reversed(base))


@pytest.fixture
def hook():
    return MockActionHook()


def test_action_hook_without_actions_returns_empty_list(hook):
    assert hook("TeSt") == []


def test_action_hook_without_actions_is_falsy(hook):
    assert not hook


def test_action_hook_with_actions_is_truthy(hook):
    hook.append_action(lowercase_action)
    assert hook


def test_action_hook_calls_action_and_returns_its_result(hook):
    hook.append_action(lowercase_action)
    assert hook("TeSt") == ["test"]


def test_action_hook_calls_multiple_actions_and_returns_their_results(hook):
    hook.append_action(lowercase_action)
    hook.append_action(uppercase_action)
    assert hook("TeSt") == ["test", "TEST"]


def test_action_hook_action_can_be_prepended_before_other_actions(hook):
    hook.prepend_action(uppercase_action)
    hook.append_action(lowercase_action)
    hook.prepend_action(reverse_action)
    assert hook("TeSt") == ["tSeT", "TEST", "test"]



from django.utils.crypto import get_random_string


def generate_theme_dirname():
    return get_random_string(8)


def upload_source_css_to(instance, filename):
    filename = add_hash_to_filename(instance.source_hash, filename)
    return "themes/%s/css/%s" % (instance.theme.dirname, filename)


def upload_build_css_to(instance, filename):
    filename = add_hash_to_filename(instance.build_hash, filename)
    return "themes/%s/css/%s" % (instance.theme.dirname, filename)


def upload_media_to(instance, filename):
    filename = add_hash_to_filename(instance.hash, filename)
    return "themes/%s/media/%s" % (instance.theme.dirname, filename)


def upload_media_thumbnail_to(instance, filename):
    return "themes/%s/media/%s" % (instance.theme.dirname, filename)


def add_hash_to_filename(hash, filename):  # pylint: disable=redefined-builtin
    if ".%s." % hash in filename:
        return filename
    extension_start = filename.rfind(".")
    return "%s.%s%s" % (filename[:extension_start], hash, filename[extension_start:])



from ..parser import parse


def test_single_line_quote(request_mock, user, snapshot):
    text = "[quote]Sit amet elit.[/quote]"
    result = parse(text, request_mock, user)
    assert snapshot == result["parsed_text"]


def test_single_line_authored_quote(request_mock, user, snapshot):
    text = '[quote="@Bob"]Sit amet elit.[/quote]'
    result = parse(text, request_mock, user)
    assert snapshot == result["parsed_text"]


def test_single_line_authored_quote_without_quotations(request_mock, user, snapshot):
    text = "[quote=@Bob]Sit amet elit.[/quote]"
    result = parse(text, request_mock, user)
    assert snapshot == result["parsed_text"]


def test_quote_can_contain_bbcode_or_markdown(request_mock, user, snapshot):
    text = "[quote]Sit **amet** [u]elit[/u].[/quote]"
    result = parse(text, request_mock, user)
    assert snapshot == result["parsed_text"]


def test_multi_line_quote(request_mock, user, snapshot):
    text = """
[quote]
Sit amet elit.

Another line.
[/quote]
"""
    result = parse(text, request_mock, user)
    assert snapshot == result["parsed_text"]


def test_quotes_can_be_nested(request_mock, user, snapshot):
    text = """
[quote]
Sit amet elit.
[quote]Nested quote[/quote]
[/quote]
"""
    result = parse(text, request_mock, user)
    assert snapshot == result["parsed_text"]


# Regression test for weird edge case in which hr gets moved outside of quote
def test_quotes_can_contain_hr_markdown(request_mock, user, snapshot):
    text = """
[quote]
Sit amet elit.
- - - - -
Another line.
[/quote]
"""
    result = parse(text, request_mock, user)
    assert snapshot == result["parsed_text"]



from django.core.exceptions import PermissionDenied
from django.utils.translation import pgettext
from rest_framework import status
from rest_framework.response import Response

from ....core.utils import format_plaintext_for_html
from ...serializers import EditSignatureSerializer
from ...signatures import is_user_signature_valid, set_user_signature


def signature_endpoint(request):
    if not request.user_acl["can_have_signature"]:
        raise PermissionDenied(
            pgettext(
                "signature api",
                "You don't have permission to change signature.",
            )
        )

    user = request.user

    if user.is_signature_locked:
        if user.signature_lock_user_message:
            reason = format_plaintext_for_html(user.signature_lock_user_message)
        else:
            reason = None

        return Response(
            {
                "detail": pgettext(
                    "signature api",
                    "Your signature is locked. You can't change it.",
                ),
                "reason": reason,
            },
            status=status.HTTP_403_FORBIDDEN,
        )

    if request.method == "POST":
        return edit_signature(request, user)

    return get_signature_options(request.settings, user)


def get_signature_options(settings, user):
    options = {
        "signature": None,
        "limit": settings.signature_length_max,
    }

    if user.signature:
        options["signature"] = {
            "plain": user.signature,
            "html": user.signature_parsed,
        }

        if not is_user_signature_valid(user):
            # pylint: disable=unsupported-assignment-operation
            options["signature"]["html"] = None

    return Response(options)


def edit_signature(request, user):
    serializer = EditSignatureSerializer(
        user, data=request.data, context={"settings": request.settings}
    )
    if serializer.is_valid():
        signature = serializer.validated_data["signature"]
        set_user_signature(request, user, request.user_acl, signature)
        user.save(
            update_fields=[
                "signature",
                "signature_parsed",
                "signature_checksum",
            ],
        )
        return get_signature_options(request.settings, user)

    return Response(
        {"detail": serializer.errors["non_field_errors"][0]},
        status=status.HTTP_400_BAD_REQUEST,
    )



from django.core.exceptions import PermissionDenied
from django.utils.translation import pgettext
from rest_framework.permissions import SAFE_METHODS, BasePermission


class IsAuthenticatedOrReadOnly(BasePermission):
    def has_permission(self, request, view):
        if request.user.is_anonymous and request.method not in SAFE_METHODS:
            raise PermissionDenied(
                pgettext("api", "This action is not available to guests.")
            )

        return True



from django.urls import reverse

from ....notifications.models import Notification


def test_notifications_read_all_api_returns_403_error_if_client_is_no_authenticated(
    db, client
):
    response = client.post(reverse("misago:apiv2:notifications-read-all"))
    assert response.status_code == 403


def test_notifications_read_all_api_marks_users_notifications_as_read(
    user, user_client
):
    notification = Notification.objects.create(user=user, verb="TEST", is_read=False)

    user.unread_notifications = 10
    user.save()

    response = user_client.post(reverse("misago:apiv2:notifications-read-all"))
    assert response.status_code == 204

    notification.refresh_from_db()
    assert notification.is_read

    # User unread notifications counter is zeroed
    user.refresh_from_db()
    assert user.unread_notifications == 0


def test_notifications_read_all_api_excludes_other_users_notifications(
    other_user, user_client
):
    notification = Notification.objects.create(
        user=other_user, verb="TEST", is_read=False
    )

    other_user.unread_notifications = 10
    other_user.save()

    response = user_client.post(reverse("misago:apiv2:notifications-read-all"))
    assert response.status_code == 204

    notification.refresh_from_db()
    assert not notification.is_read

    other_user.refresh_from_db()
    assert other_user.unread_notifications == 10


def test_notifications_read_all_api_works_when_user_has_no_unread_notifications(
    user, user_client
):
    notification = Notification.objects.create(user=user, verb="TESt", is_read=True)

    user.unread_notifications = 0
    user.save()

    response = user_client.post(reverse("misago:apiv2:notifications-read-all"))
    assert response.status_code == 204

    notification.refresh_from_db()
    assert notification.is_read

    # User unread notifications counter is zeroed
    user.refresh_from_db()
    assert user.unread_notifications == 0


def test_notifications_read_all_api_works_when_user_has_no_notifications(
    user, user_client
):
    user.unread_notifications = 0
    user.save()

    response = user_client.post(reverse("misago:apiv2:notifications-read-all"))
    assert response.status_code == 204

    # User unread notifications counter is zeroed
    user.refresh_from_db()
    assert user.unread_notifications == 0



from io import StringIO

from django.contrib.auth import get_user_model
from django.core.management import call_command
from django.test import TestCase

from ...conf.test import override_dynamic_settings
from ..management.commands import deletemarkedusers
from ..models import DeletedUser
from ..test import create_test_user

User = get_user_model()


class DeleteMarkedUsersTests(TestCase):
    def setUp(self):
        self.user = create_test_user("User", "user@example.com")
        self.user.mark_for_delete()

    def test_delete_marked_user(self):
        """deletes marked user"""
        out = StringIO()
        call_command(deletemarkedusers.Command(), stdout=out)
        command_output = out.getvalue().splitlines()[0].strip()

        self.assertEqual(command_output, "Deleted users: 1")

        with self.assertRaises(User.DoesNotExist):
            self.user.refresh_from_db()

    def test_marked_user_deletion_is_recorded(self):
        out = StringIO()
        call_command(deletemarkedusers.Command(), stdout=out)
        command_output = out.getvalue().splitlines()[0].strip()

        self.assertEqual(command_output, "Deleted users: 1")

        DeletedUser.objects.get(deleted_by=DeletedUser.DELETED_BY_SELF)

    @override_dynamic_settings(allow_delete_own_account=False)
    def test_delete_disabled(self):
        """deletion respects user decision even if configuration has changed"""
        out = StringIO()
        call_command(deletemarkedusers.Command(), stdout=out)
        command_output = out.getvalue().splitlines()[0].strip()

        self.assertEqual(command_output, "Deleted users: 1")

        with self.assertRaises(User.DoesNotExist):
            self.user.refresh_from_db()

    def test_delete_not_marked(self):
        """user has to be marked to be deletable"""
        self.user.is_deleting_account = False
        self.user.save()

        out = StringIO()
        call_command(deletemarkedusers.Command(), stdout=out)
        command_output = out.getvalue().splitlines()[0].strip()

        self.assertEqual(command_output, "Deleted users: 0")

        self.user.refresh_from_db()

    def test_delete_is_staff(self):
        """staff users are extempt from deletion"""
        self.user.is_staff = True
        self.user.save()

        out = StringIO()
        call_command(deletemarkedusers.Command(), stdout=out)
        command_output = out.getvalue().splitlines()[0].strip()

        self.assertEqual(command_output, "Deleted users: 0")

        self.user.refresh_from_db()

    def test_delete_superuser(self):
        """superusers are extempt from deletion"""
        self.user.is_superuser = True
        self.user.save()

        out = StringIO()
        call_command(deletemarkedusers.Command(), stdout=out)
        command_output = out.getvalue().splitlines()[0].strip()

        self.assertEqual(command_output, "Deleted users: 0")

        self.user.refresh_from_db()



from math import ceil

from django.urls import reverse
from rest_framework import serializers

from ...categories.serializers import CategorySerializer
from ...core.serializers import MutableFields
from ...notifications.threads import ThreadNotifications
from ..models import Thread
from .poll import PollSerializer
from .threadparticipant import ThreadParticipantSerializer

__all__ = ["ThreadSerializer", "PrivateThreadSerializer", "ThreadsListSerializer"]

BasicCategorySerializer = CategorySerializer.subset_fields(
    "id",
    "parent",
    "name",
    "short_name",
    "color",
    "description",
    "is_closed",
    "css_class",
    "level",
    "lft",
    "rght",
    "is_read",
    "url",
)


class ThreadSerializer(serializers.ModelSerializer, MutableFields):
    category = BasicCategorySerializer(many=False, read_only=True)

    acl = serializers.SerializerMethodField()
    has_unapproved_posts = serializers.SerializerMethodField()
    is_new = serializers.SerializerMethodField()
    is_read = serializers.SerializerMethodField()
    path = BasicCategorySerializer(many=True, read_only=True)
    poll = PollSerializer(many=False, read_only=True)
    pages = serializers.SerializerMethodField()
    best_answer = serializers.PrimaryKeyRelatedField(read_only=True)
    best_answer_marked_by = serializers.PrimaryKeyRelatedField(read_only=True)
    notifications = serializers.SerializerMethodField()
    starter = serializers.SerializerMethodField()
    last_poster = serializers.SerializerMethodField()

    api = serializers.SerializerMethodField()
    url = serializers.SerializerMethodField()

    class Meta:
        model = Thread
        fields = [
            "id",
            "category",
            "title",
            "replies",
            "has_unapproved_posts",
            "started_on",
            "starter_name",
            "last_post_on",
            "last_post_is_event",
            "last_post",
            "last_poster_name",
            "is_unapproved",
            "is_hidden",
            "is_closed",
            "weight",
            "best_answer",
            "best_answer_is_protected",
            "best_answer_marked_on",
            "best_answer_marked_by",
            "best_answer_marked_by_name",
            "best_answer_marked_by_slug",
            "acl",
            "is_new",
            "is_read",
            "path",
            "poll",
            "notifications",
            "starter",
            "last_poster",
            "pages",
            "api",
            "url",
        ]

    def get_acl(self, obj):
        try:
            return obj.acl
        except AttributeError:
            return {}

    def get_has_unapproved_posts(self, obj):
        try:
            acl = obj.acl
        except AttributeError:
            return False
        return acl.get("can_approve") and obj.has_unapproved_posts

    def get_is_new(self, obj):
        try:
            return obj.is_new
        except AttributeError:
            return None

    def get_is_read(self, obj):
        try:
            return obj.is_read
        except AttributeError:
            return None

    def get_participants(self, obj):
        return ThreadParticipantSerializer(obj.participants_list, many=True).data

    def get_notifications(self, obj):
        if self.context:
            watched_thread = self.context.get("watched_thread")
            if watched_thread:
                if watched_thread.send_emails:
                    return ThreadNotifications.SITE_AND_EMAIL
                return ThreadNotifications.SITE_ONLY

            watched_threads = self.context.get("watched_threads")
            if watched_threads:
                return watched_threads.get(obj.id)

        return None

    def get_starter(self, obj):
        if obj.starter_id:
            return {
                "id": obj.starter_id,
                "username": obj.starter.username,
                "real_name": obj.starter.get_real_name(),
                "avatars": obj.starter.avatars,
            }

    def get_last_poster(self, obj):
        if obj.last_poster_id:
            return {
                "id": obj.last_poster_id,
                "username": obj.last_poster.username,
                "real_name": obj.last_poster.get_real_name(),
                "avatars": obj.last_poster.avatars,
            }

    def get_pages(self, obj):
        settings = self.context["settings"]

        posts_per_page = settings.posts_per_page - 1
        posts_per_page_orphans = settings.posts_per_page_orphans

        if posts_per_page_orphans:
            posts_per_page_orphans += 1

        total_posts = obj.replies + 1
        if total_posts <= posts_per_page + posts_per_page_orphans:
            return 1

        hits = total_posts - posts_per_page_orphans
        return ceil(hits / posts_per_page)

    def get_api(self, obj):
        return {
            "index": obj.get_api_url(),
            "editor": obj.get_editor_api_url(),
            "merge": obj.get_merge_api_url(),
            "poll": obj.get_poll_api_url(),
            "watch": obj.get_watch_api_url(),
            "posts": {
                "index": obj.get_posts_api_url(),
                "merge": obj.get_post_merge_api_url(),
                "move": obj.get_post_move_api_url(),
                "split": obj.get_post_split_api_url(),
            },
        }

    def get_url(self, obj):
        return {
            "index": obj.get_absolute_url(),
            "new_post": obj.get_new_post_url(),
            "last_post": obj.get_last_post_url(),
            "best_answer": obj.get_best_answer_url(),
            "unapproved_post": obj.get_unapproved_post_url(),
            "starter": self.get_starter_url(obj),
            "last_poster": self.get_last_poster_url(obj),
        }

    def get_starter_url(self, obj):
        if obj.starter_id:
            return reverse(
                "misago:user", kwargs={"slug": obj.starter_slug, "pk": obj.starter_id}
            )

    def get_last_poster_url(self, obj):
        if obj.last_poster_id:
            return reverse(
                "misago:user",
                kwargs={"slug": obj.last_poster_slug, "pk": obj.last_poster_id},
            )


class PrivateThreadSerializer(ThreadSerializer):
    participants = serializers.SerializerMethodField()

    class Meta:
        model = Thread
        fields = ThreadSerializer.Meta.fields + ["participants"]


class ThreadsListSerializer(ThreadSerializer):
    category = serializers.PrimaryKeyRelatedField(read_only=True)
    last_post = serializers.PrimaryKeyRelatedField(read_only=True)

    class Meta:
        model = Thread
        fields = ThreadSerializer.Meta.fields + ["has_poll"]


ThreadsListSerializer = ThreadsListSerializer.exclude_fields("path", "poll")



from ..models import Notification


def test_notification_actor_name_is_anonymized(user, other_user):
    notification = Notification.objects.create(
        user=user,
        actor=other_user,
        actor_name=other_user.username,
        verb="TEST",
    )

    other_user.anonymize_data("Deleted")

    notification.refresh_from_db()
    assert notification.actor_name == "Deleted"



from django.apps import AppConfig


class MisagoNotificationsConfig(AppConfig):
    name = "misago.notifications"
    label = "misago_notifications"
    verbose_name = "Misago Notifications"

    def ready(self):
        from . import signals as _  # noqa: F401
